################################################################################
#
# tgdb - A graphical frontend to gdb, the GNU debugger.
#	 Based on Tcl/Tk, TclX, TkSteal, Expect and BLT.
#
# (c) 1994-99 HighTec EDV-Systeme GmbH
#             Feldmannstr. 98
#             66119 Saarbruecken, Germany
#             Email: tgdb@HighTec.saarlink.de
#
# *** ALL RIGHTS RESERVED ***
#
################################################################################

set tgdb_version 8.0
wm withdraw .

################################################################################
#
# load additional libraries (aka Tcl/Tk extensions :) into this interpreter
#
################################################################################

if { [info exists env(TGDB_BINPATH)] } {
  source $env(TGDB_BINPATH)/loadlibs.tcl
} else {
  catch {source $env(TGDB_PATH)/bin/$tcl_platform(os)/loadlibs.tcl}
}

#
# die if TclX isn't available (BLT, TkSteal and Expect are optional now)
#
if { ![info exists tclx_library] } {
  catch {bell}
  puts "\nFATAL: TclX not available; exiting.\n"
  exit 1
}
if { [info exists blt_version] && ($blt_version >= 2.4) } {
  namespace import blt::*
}

################################################################################
#
# tgdb/gdb interface stuff
#
################################################################################

if { [info exists env(TGDB_SOCKET)] } {
  set use_socket 1
} else {
  set use_socket 0
}
if { "[info commands expect]" == "" } {
  set use_expect 0
} else {
  set use_expect 1
}
if { ![info exists env(TGDB_REMOTE)] && !$use_socket } {
  set use_remote 0
} else {
  set use_expect 0
  set use_remote 1
  set remote_readline 0

  if { $use_socket } {
    proc srv_accept_client { sock ip_addr port } {
      global srv_sock gdb_desc

      close $srv_sock
      fconfigure $sock -blocking 0 -buffering none
      fileevent $sock readable handle_remote_input
      set gdb_desc $sock
    }

    proc srv_open_socket {} {
      global srv_sock env

      for { set port 5555 } { $port < 6000 } { incr port } {
        if { ![catch {set srv_sock [socket -server srv_accept_client $port]}]} {
	  set env(TGDB_SOCKETPORT) $port
	  return
	}
      }
      puts "Couldn't find an unused port to open a socket."
      exit 1
    }
  }

  proc handle_remote_input {} {
    global gdb_desc remote_result remote_ready remote_readline prompt PS2
    global use_socket QueryResult Tgdb_option

    set input ""
    if { [catch {set input [read $gdb_desc]}] || [eof $gdb_desc] } {
      catch {close $gdb_desc}
      exit_tgdb
      return
    }
    append remote_result $input
    #
    # check if input contains PXMon notifications; this has to be done
    # first, 'cos it can potentially contain a prompt, too!
    #
    if { [set pos [string first "\n\x1a\x1aInterrupt\n" $remote_result]] > -1 } {
      if { $Tgdb_option(TraceGDB) } {
	puts $std_out "RECV <\n\x1a\x1aInterrupt\n>"
      }
      if { $pos == 0 } {
	set remote_result [crange $remote_result 13 end]
      } else {
        set first [crange $remote_result 0 $pos-1]
	set $first[crange $remote_result $pos+13 end]
      }
      show_status "Received target messages (use \"ptm\" to process them)." \
        steady
      bell
    }
    #
    # check if we've received the prompt
    #
    set pl [clength $prompt]
    set rl [clength $remote_result]
    if { ($rl>=$pl) && [cequal [crange $remote_result $rl-$pl end] $prompt] } {
      if { $rl > $pl } {
	set remote_result [crange $remote_result 0 $rl-$pl-1]
      } else {
	set remote_result ""
      }
      set remote_ready 1
      return
    }
    #
    # check if it contains a query
    #
    set lines [split $remote_result \n]
    set nlines [llength $lines]
    set lastlines [join [lrange $lines [expr $nlines - 2] end] \n]
    if { [regexp ".*\\? \\\(y or n\\\) \$" $lastlines {}] } {
      if { $nlines > 2 } {
	set remote_result [crange $remote_result 0 end-[clength $lastlines]]
      } else {
	set remote_result ""
      }
      if { [yes_no_box $lastlines] } {
	append QueryResult y
	if { $use_socket } {
	  puts $gdb_desc y\04
	} else {
	  puts $gdb_desc y
	}
      } else {
	append QueryResult n
	if { $use_socket } {
	  puts $gdb_desc n\04
	} else {
	  puts $gdb_desc n
	}
      }
      return
    }
    #
    # check if it contains readline stuff (cf. "comm", "def", "doc")
    #
    if { $remote_readline } {
      if {   [regexp ".*End with a line saying just \"end\"\.\n>\$" $lastlines]
	  || [regexp {^>$} $lastlines] } {
        set remote_ready 1
        return
      }
    }
    #
    # check if it contains a C++ breakpoint selection menu
    #
    if { [regexp -indices "(\\\[(\[0-9\])+\\\]\[^\n\]*\n)+$PS2 " \
	   $remote_result selidx] } {
      if { ![lindex $selidx 0] } {
        if { $use_socket } {
	  puts $gdb_desc [choose_bpt $remote_result]\04
	} else {
	  puts $gdb_desc [choose_bpt $remote_result]
	}
	set remote_result ""
      } else {
	set menu [crange $remote_result [lindex $selidx 0] end]
        if { $use_socket } {
	  puts $gdb_desc [choose_bpt $menu]\04
	} else {
	  puts $gdb_desc [choose_bpt $menu]
	}
	set remote_result [crange $remote_result 0 [lindex $selidx 0]-1]
      }
    }
    #
    # check if a new thread was created
    #
    if { [regexp -indices "\n\\\[New (\[^\]\]*)\\\]\n" $remote_result \
           {} selidx] } {
      set task [eval string range {$remote_result} $selidx]
      .f5.text insert end "New task $task\n"
      .f5.text yview -pickplace end
      gdbtk_new_task $task
      regsub "\n\\\[New (\[^\]\]*)\\\]\n" $remote_result "" remote_result
    }
    #
    # check if switching to another thread is required
    #
    if { [regexp -indices "\n\\\[Switching to (\[^\]\]*)\\\]\n" $remote_result \
	   {} selidx] } {
      set task [eval string range {$remote_result} $selidx]
      gdbtk_switch_to_task $task
      regsub "\n\\\[Switching to (\[^\]\]*)\\\]\n" $remote_result "" \
        remote_result
    }
  }

  proc strip_prompt { lines } {
    global prompt

    set pl [clength $prompt]
    set ll [clength $lines]
    if { [cequal [crange $lines $ll-$pl end] $prompt] } {
      return [crange $lines 0 $ll-$pl-1]
    } else {
      return $lines
    }
  }
}

if {   [info exists env(GDBTK_FILENAME)]
    && [regexp {tgdb.*} [file tail $env(GDBTK_FILENAME)]]
    && ![regexp {.*wish$} [info nameofexecutable]] } {
  set use_expect 0
  set use_remote 0
}

if { $use_remote } {
  proc gdb_cmd { cmd } {
    global gdb_desc remote_result remote_ready use_socket

    set remote_result ""
    catch {unset remote_ready}
    if { $use_socket } {
      append cmd \04
    }
    if { [catch {puts $gdb_desc $cmd}] } {
      exit_tgdb
      return -code error
    }
    vwait remote_ready
    return $remote_result
  }
}

if {   [info exists env(TGDB_NODEBUGEEWINDOW)]
    || ("[info commands xaccess]" == "") } {
  set use_dbgwin 0
} else {
  if { [info exists tcl_platform] && ("$tcl_platform(platform)" == "unix") } {
    set use_dbgwin 1
  } else {
    set use_dbgwin 0
  }
}

################################################################################
#
# Tk compatibility stuff (3.x -> 4.x)
#
################################################################################

if { $tk_version > 3.6 } {
  #
  # newer versions of TkSteal use tkSteal instead of tksteal :-)
  #
  catch {tksteal} ;# force auto-loading
  catch {tkSteal}
  if { "[info comm tkSteal]" == "tkSteal" } {
    rename tkSteal tksteal
  }
  #
  # define wrapper functions for interactive widgets
  #
  set code {
    rename widget __widget
    proc widget { args } {
      set w [eval __widget $args]
      bindtags $w "$w . all"
      return $w
    }
  }
  foreach widget "text" {
    regsub -all "widget" $code $widget wrapcode
    uplevel #0 $wrapcode
  }
  unset code wrapcode
  #
  # override some 4.x bindings
  #
  bind Entry <ButtonRelease-2> { }
  bind Text <ButtonRelease-2> { }
  bind Text <BackSpace> { tk_textBackspace %W }
  bind Button <space> { }
  bind Button <Return> { }
  bind Checkbutton <space> { }
  bind Checkbutton <Return> { }
  bind Radiobutton <space> { }
  bind Radiobutton <Return> { }
  bind all <Tab> { }
  bind all <Shift-Tab> { }
  #
  # define Tk compatibility procs
  #
  proc tk_textBackspace { w } {
    if { [$w compare insert != 1.0] } {
      $w delete insert-1c
      $w see insert
    }
  }
  proc tk_entrySeeCaret { w } {
    tkEntrySeeInsert $w
  }
# if { "$tcl_platform(platform)" != "unix" } {
#   # FIXME: will this bug disappear in 4.1a3?
#   rename tkwait __tkwait
#   proc tkwait { action args } {
#     if { [string first $action visibility] != 0 } {
#	return [eval __tkwait $action $args]
#     }
#   }
# }
  proc tk_textSelectTo {w index} {
    global tk_priv

    if [catch {$w index anchor}] {
      $w mark set anchor $index
    }
    case $tk_priv(selectMode) {
      char { if [$w compare $index < anchor] {
               set first $index
               set last anchor
             } else {
               set first anchor
               set last [$w index $index+1c]
             }
           }
      word { if [$w compare $index < anchor] {
               set first [$w index "$index wordstart"]
               set last [$w index "anchor wordend"]
             } else {
               set first [$w index "anchor wordstart"]
               set last [$w index "$index wordend"]
             }
           }
      line { if [$w compare $index < anchor] {
               set first [$w index "$index linestart"]
               set last [$w index "anchor lineend + 1c"]
             } else {
               set first [$w index "anchor linestart"]
               set last [$w index "$index lineend + 1c"]
             }
           }
    }
    $w tag remove sel 0.0 $first
    $w tag add sel $first $last
    $w tag remove sel $last end
  }
  #
  # backward compatibility proc for focus
  #
  rename toplevel __toplevel
  proc toplevel { args } {
    if { [catch {set result [eval __toplevel $args]} error] } {
      return -code error -errorinfo $error
    } else {
      frame [lindex $args 0].none
      return $result
    }
  }

  frame .none
  #focus .none
  rename focus __focus
  proc focus { args } {
    if { [cequal [file extension $args] .none] } {
      set result [__focus]
      if { [catch {__focus [winfo toplevel $result].none}] } {
        __focus .none
      }
      return ""
    } elseif { "$args" == "" } {
      set result [__focus]
      if { [cequal [file extension $result] .none] } {
        set result .none
      }
    } elseif { [lsearch $args -f*] > -1 } {
      set result [eval __focus $args]
    } else {
      set result [eval __focus -force $args]
    }
    if { ("$args" == "") && ($result == "") } {
      return .none
    } else {
      return $result
    }
  }
  #
  # miscellaneous stuff
  #
  # bell is defined by both, BLT 2.x and Tk 4.x
  rename bell tk_bell
  #
  # get rid of focus borders and unwanted traversal
  #
  option add *highlightThickness 0
  option add *Frame.takeFocus 0
  # set background color for entries and texts to that of labels
  if { [cequal $tcl_platform(platform) windows] } {
    label .l
    option add *Text.background [.l cget -bg]
    option add *Entry.background [.l cget -bg]
    destroy .l
  }
}

################################################################################
# session management
################################################################################

if { [info exists tcl_platform] && ("$tcl_platform(platform)" == "unix") } {
  catch {
    # drop a note to the session manager
    set hostname [exec hostname]
    if { [set domain [exec domainname]] != "" } {
      append hostname "." $domain
    }
    wm client . $hostname
    wm application . $argv
    unset hostname domain
  }
  set top_left_arrow top_left_arrow
}
if { [info exists tcl_platform] && ("$tcl_platform(platform)" == "windows") } {
  set top_left_arrow arrow
}

################################################################################
# BLT stuff
################################################################################

if { "[info commands htext]" == "htext" } {
  # tgdb 1.x has used BLT's hypertext widget, but it's now replaced by
  # the improved text widget of the Tk 4.x series
  rename htext {}
}
if { "[info commands barchart]" == "barchart" } {
  rename barchart blt_barchart
}
if { "[info commands busy]" == "busy" } {
  rename busy blt_busy
}

if { $tk_version < 4.0 } {
  catch {selection clear .}
}
tk appname tgdb
set appname [winfo name .]
set Tgdb_cwd [pwd]
if { $use_expect } {
  log_user 0
}
set std_out stdout
# wanna debug? ;^)
proc tron {} {
  global std_out

  if { [cequal $std_out stdout] } {
    set std_out [open dump.out w]
    cmdtrace on $std_out
  }
}
proc troff {} {
  global std_out

  if { ![cequal $std_out stdout] } {
    cmdtrace off
    catch {close $std_out}
    catch {set std_out stdout}
  }
}
if 0 tron

################################################################################
#
# set default values for debugger, prompt, etc.
#
################################################################################

if { [catch {set debugger $env(TGDB_DEBUGGER)}] } {
  set debugger gdb
}
# see if this is gonna be a remote session
if { [cequal [lindex $debugger 0] rsh] } {
  set dbgrsh [lrange $debugger 0 [expr [llength $debugger] - 2]]
  set debugger [lindex $debugger [expr [llength $debugger] - 1]]
} else {
  set dbgrsh ""
}
if { "$debugger" == "gdb166" } {
  set use_dbgwin 0
}
if { [catch {set prompt $env(TGDB_PROMPT)}] } {
  set prompt "($debugger) "
}
if { [catch {set PS2 $env(PS2)}] || ("$PS2" == "") } {
  set PS2 [set env(PS2) ">"]
} else { 
  regsub -all {([][^(){}\|.?+*$])} $PS2 {\\\1} PS2                       
}
if { [catch {set tgdb_path $env(TGDB_PATH)}] } {
  set tgdb_path [pwd]
}
lappend auto_path $tgdb_path
if { [info exists tcl_platform] && ![cequal $tcl_platform(platform) windows] } {
  append env(PATH) ":$tgdb_path"
}
if { [info exists env(TGDB_SMALLICONS)] && ($env(TGDB_SMALLICONS) == "1") } {
  set bitmap_path "$tgdb_path/icon16"
} else {
  set bitmap_path "$tgdb_path/icon32"
}
if { [catch {set tgdb_appdef_path $env(TGDB_APPDEF_PATH)}] } {
  set tgdb_appdef_path [pwd]
}
if { $prompt == "(gdb16x) " } {
  set gdb16x 1
} else {
  set gdb16x 0
}
if { [info exists tcl_platform] } {
  if { [cequal $tcl_platform(platform) unix] } {
    set gdb_initfile "./.gdbinit"
  } else {
    set gdb_initfile "./gdb.ini"
  }
} else {
  set gdb_initfile "./.gdbinit"
}

################################################################################
# examine the argument(s) which we should pass to gdb; do this now,
# so that we can quickly exit if something's wrong with them
################################################################################

process_gdb_args

################################################################################
#
# load X application default database
#
################################################################################

if { [winfo depth .] > 1 } {
  set colormodel color
} else {
  set colormodel monochrome
}
proc read_option_file { filename {prio interactive} } {
  global appname colormodel

  # read an option file (replace application name if necessary)
  if { $colormodel == "color" } {
    if { $appname == "tgdb" } {
      option readfile $filename $prio
    } else {
      for_file line $filename {
        if { [regsub {^[ 	]*tgdb} $line $appname line] } {
	  if { [set seppos [string first ":" $line]] == -1 } continue
	  set option [string trim [crange $line 0 $seppos-1]]
	  set value [string trim [crange $line $seppos+1 end]]
	  option add $option $value $prio
        }
      }
    }
  } else {
    for_file line $filename {
      if { [regsub {^[ 	]*tgdb} $line $appname line] } {
	if { [set seppos [string first ":" $line]] == -1 } continue
	set option [string trim [crange $line 0 $seppos-1]]
	set value [string trim [crange $line $seppos+1 end]]
	if { [string first foreground $option] != -1 } {
	  set value black
	} elseif { [string first background $option] != -1 } {
	  set value white
	}
	option add $option $value $prio
      }
    }
  }
}
if { $tk_version < 8.0 } {
  set Tgdb_big settings/$tcl_platform(platform)/Tgdb-big.pre8
  set Tgdb_small settings/$tcl_platform(platform)/Tgdb-small.pre8
} else {
  set Tgdb_big settings/$tcl_platform(platform)/Tgdb-big
  set Tgdb_small settings/$tcl_platform(platform)/Tgdb-small
}
if { [info exists env(TGDB_SMALLICONS)] && ($env(TGDB_SMALLICONS) == "1") } {
  if { [file readable "./.Xdefaults-small"] } {
    read_option_file "./.Xdefaults-small"
  } elseif { [file readable "$tgdb_appdef_path/$Tgdb_small"] } {
    read_option_file "$tgdb_appdef_path/$Tgdb_small"
  } elseif { [file readable "~/.Xdefaults"] } {
    read_option_file "~/.Xdefaults"
  }
} else {
  if { [file readable "./.Xdefaults-big"] } {
    read_option_file "./.Xdefaults-big"
  } elseif { [file readable "$tgdb_appdef_path/$Tgdb_big"] } {
    read_option_file "$tgdb_appdef_path/$Tgdb_big"
  } elseif { [file readable "~/.Xdefaults"] } {
    read_option_file "~/.Xdefaults"
  }
}

################################################################################
#
# say hello to the user
#
################################################################################

intro

################################################################################
#
# create main window's widgets
#
################################################################################

frame .f0 -relief raised -border 2
frame .f0.f0 -relief raised -border 2
#-------------------------------------------------------------------------------
menubutton .f0.f0.file -text {File } -relief flat -menu .f0.f0.file.m
menu .f0.f0.file.m
menu .f0.f0.file.m.visit
.f0.f0.file.m add command -label { Load program } -command {file_command}
.f0.f0.file.m add command -label { Load executable file } \
  -command {exec_command}
.f0.f0.file.m add command -label { Load symbol file } -command {symbol_command}
.f0.f0.file.m add command -label { Load core file } -command {core_command}
.f0.f0.file.m add command -label { Load gdb source file } \
  -command {read_source_file}
.f0.f0.file.m add command -label { Load text file } -command {load_text_command}
.f0.f0.file.m add cascade -label { Visit } -menu .f0.f0.file.m.visit \
  -state disabled
.f0.f0.file.m add command -label { Save gdb window } -command {save_gdb_window}
.f0.f0.file.m add command -label { Save gdb init file } \
  -command {write_init_file [FileSelectBox -title "Select init file" -perm w]}
.f0.f0.file.m add command -label { Add source directory } \
  -command {add_source_dir}
.f0.f0.file.m add command -label { Change working directory } \
  -command {chg_directory}
.f0.f0.file.m add separator
.f0.f0.file.m add command -label { Quit } -command {exit_tgdb} \
  -accelerator {Ctrl-c}
#-------------------------------------------------------------------------------
menubutton .f0.f0.options -text { Options } -relief flat -menu .f0.f0.options.m
menu .f0.f0.options.m
.f0.f0.options.m add checkbutton -label { Visual bell } \
  -onvalue visual -offvalue audible -variable {Tgdb_option(bell)}
set RaiseLabel " Auto-raise covered windows "
.f0.f0.options.m add checkbutton -label $RaiseLabel \
  -onvalue 1 -offvalue 0 -variable {Tgdb_option(RaiseWindows)} \
  -command raise_windows -accelerator {Alt-Esc}
if { $use_dbgwin } {
  .f0.f0.options.m add checkbutton -label { Auto-raise debugee window } \
    -onvalue 1 -offvalue 0 -variable {Tgdb_option(RaiseDbgWindow)}
  .f0.f0.options.m add checkbutton -label { Warp pointer to debugee window } \
    -onvalue 1 -offvalue 0 -variable {Tgdb_option(WarpToDbgWindow)}
}
.f0.f0.options.m add checkbutton -label { Make gdb window read-only } \
  -onvalue 1 -offvalue 0 -variable {Tgdb_option(gdbRO)}
.f0.f0.options.m add checkbutton -label { Trace tgdb/gdb dialog } \
  -onvalue 1 -offvalue 0 -variable {Tgdb_option(TraceGDB)}
.f0.f0.options.m add checkbutton -label { Update breakpoints on each stop } \
  -onvalue 1 -offvalue 0 -variable {Tgdb_option(UpdateBpts)}
.f0.f0.options.m add checkbutton -label { Enable class breakpoints (C++) } \
  -onvalue 1 -offvalue 0 -variable {Tgdb_option(ClassKludge)}
.f0.f0.options.m add checkbutton -label { Auto-save & load command history } \
  -onvalue 1 -offvalue 0 -variable {Tgdb_option(SaveHistory)}
.f0.f0.options.m add checkbutton -label { cd to program path after loading } \
  -onvalue 1 -offvalue 0 -variable {Tgdb_option(ProgPath)}
.f0.f0.options.m add checkbutton -label { Remove CRs (^M) from source files } \
  -onvalue 1 -offvalue 0 -variable {Tgdb_option(RemCRs)}
.f0.f0.options.m add separator
.f0.f0.options.m add checkbutton -label { Save preferences on exit } \
  -onvalue 1 -offvalue 0 -variable {Tgdb_option(SaveOptions)}
#-------------------------------------------------------------------------------
menubutton .f0.f0.running -text { Running } -relief flat -menu .f0.f0.running.m
menu .f0.f0.running.m
.f0.f0.running.m add command -label { Run / restart program } \
  -accelerator {Alt-z} -command {$Tgdb_cmd(run)}
.f0.f0.running.m add command -label { Continue } -accelerator {c<Return>} \
  -command {$Tgdb_cmd(continue)}
.f0.f0.running.m add command -label { Next line } -accelerator {n<Return>} \
  -command {$Tgdb_cmd(next)}
.f0.f0.running.m add command -label { Step line } -accelerator {s<Return>} \
  -command {$Tgdb_cmd(step)}
.f0.f0.running.m add command -label { Next instruction } \
  -accelerator {ni<Return>} -command {$Tgdb_cmd(nexti)}
.f0.f0.running.m add command -label { Step instruction } \
  -accelerator {si<Return>} -command {$Tgdb_cmd(stepi)}
.f0.f0.running.m add command -label { Finish } -accelerator {fin<Return>} \
  -command {$Tgdb_cmd(finish)}
.f0.f0.running.m add command -label { Return } -accelerator {ret<Return>} \
  -command {$Tgdb_cmd(return)}
.f0.f0.running.m add separator
.f0.f0.running.m add command -label { Update Threads } -state disabled \
  -command update_tasklist
#-------------------------------------------------------------------------------
menubutton .f0.f0.windows -text { Windows } -relief flat -menu .f0.f0.windows.m
menu .f0.f0.windows.m
.f0.f0.windows.m add command -label { Tgdb } -accelerator {Alt-t} \
  -command {raise_main_window}
if { $use_dbgwin } {
.f0.f0.windows.m add command -label { Program (debugee) } -accelerator {Alt-p} \
  -command {raise_debug_window}
}
.f0.f0.windows.m add command -label { Stack } -accelerator {Alt-s} \
  -command {create_stack_window}
.f0.f0.windows.m add command -label { CPU registers } -accelerator {Alt-r} \
  -command {create_cpu_window}
.f0.f0.windows.m add command -label { Watches } -accelerator {Alt-w} \
  -command {create_disp_window}
.f0.f0.windows.m add command -label { Memory dump } -accelerator {Alt-d} \
  -command {create_memory_window}
.f0.f0.windows.m add command -label { Assembly dump } -accelerator {Alt-a} \
  -command {create_assembly_window}
if { "[info commands xaccess]" == "xaccess" } {
  .f0.f0.windows.m add command -label { Xterm (shell) } -accelerator {Alt-x} \
    -command {create_shell_window}
}
.f0.f0.windows.m add command -label { PXMon } -accelerator {Alt-o} \
  -command {create_pxmon_window}
.f0.f0.windows.m add separator
.f0.f0.windows.m add command -label { Cycle windows } -accelerator {Alt-c} \
  -command {toggle_windows}
#-------------------------------------------------------------------------------
menubutton .f0.f0.cmds -text { Commands } -relief flat -menu .f0.f0.cmds.m
menu .f0.f0.cmds.m
.f0.f0.cmds.m add command -label { Attach commands to a breakpoint } \
  -command {do_edit commands}
.f0.f0.cmds.m add command -label { Define a user command } \
  -command {do_edit define}
.f0.f0.cmds.m add command -label { Document a user command } \
  -command {do_edit document}
.f0.f0.cmds.m add separator
.f0.f0.cmds.m add command -label { Clear command history } \
  -command {clear_command_history}
#-------------------------------------------------------------------------------
menubutton .f0.f0.thrds -text { Threads } -relief flat -menu .f0.f0.thrds.m
menu .f0.f0.thrds.m -tearoff 0
.f0.f0.thrds.m add radiobutton -label { Global } -variable M_thread \
  -value 0 -command { t_thread 0 }
#-------------------------------------------------------------------------------
#menubutton .f0.f0.info -text { Info } -relief flat -menu .f0.f0.info.m
#menu .f0.f0.info.m
#.f0.f0.info.m add command -label { Variables } -state disabled
#.f0.f0.info.m add command -label { Functions } -state disabled
#.f0.f0.info.m add command -label { Types } -state disabled
#.f0.f0.info.m add command -label { Program } -state disabled
#.f0.f0.info.m add command -label { Sources } -state disabled
#-------------------------------------------------------------------------------
menubutton .f0.f0.help -text { Help} -relief flat -menu .f0.f0.help.m
menu .f0.f0.help.m
.f0.f0.help.m add command -label { About tgdb } -accelerator {Alt-i} \
  -command help_about
.f0.f0.help.m add command -label { About HighTec } \
  -command help_about_hightec
.f0.f0.help.m add command -label { tgdb } -accelerator {Alt-h} \
  -command create_tgdb_help_window ;#-state disabled
.f0.f0.help.m add command -label { gdb commands } -accelerator {h<Return>} \
  -command show_help

################################################################################
frame .f1 -relief raised -border 2
entry .f1.status -state disabled -exportselection 0 -textvariable TextStatus
################################################################################
frame .title -relief raised -border 2
label .title.title -textvariable ThisFile
frame .title.f1 -relief raised -border 2
label .title.f1.line -text {Line: } -relief flat
entry .title.f1.lentry -textvariable Line -width 5 -relief flat \
  -state disabled
################################################################################
frame .f2 -relief raised -border 4
label .f2.lbl1 -text {Search} -relief flat
label .f2.lbl2 -text {: } -relief flat
entry .f2.search -textvariable SearchString -relief flat
frame .f2.f
button .f2.f.up -height 4 -bitmap @$bitmap_path/arrow_up.xbm \
  -command {search_string up}
button .f2.f.dn -height 4 -bitmap @$bitmap_path/arrow_dn.xbm \
  -command {search_string down}
################################################################################
frame .f3 -relief flat
text .f3.text -exportselection 1 -setgrid 1 -yscrollcommand {.f3.scroll set} \
  -state disabled -cursor $top_left_arrow
scrollbar .f3.scroll -orient vertical -command {text_scroll}
################################################################################
frame .f4 -relief raised -border 2
button .f4.list -bitmap @$bitmap_path/list.xbm -command { list_sel_file }
button .f4.list_l -bitmap @$bitmap_path/list_l.xbm -command { list_last_file }
button .f4.list_d -bitmap @$bitmap_path/list_d.xbm -command { list_disassembly }
button .f4.list_f -bitmap @$bitmap_path/list_f.xbm -command { sel_frame frame }
button .f4.frame_dn -bitmap @$bitmap_path/frame_dn.xbm -command {sel_frame down}
button .f4.frame_0 -bitmap @$bitmap_path/frame_0.xbm \
  -command { sel_frame frame 0 }
button .f4.frame_up -bitmap @$bitmap_path/frame_up.xbm -command { sel_frame up }
button .f4.frame_in -bitmap @$bitmap_path/frame_in.xbm \
  -command { $Tgdb_cmd([list info frame]) }
button .f4.whatis -bitmap @$bitmap_path/whatis.xbm -command {
  if { [cequal [get_selection] ""] } {
    show_status "No selection." steady
  } else {
    $Tgdb_cmd(whatis) [get_selection]
  }
}
button .f4.print -bitmap @$bitmap_path/print.xbm -command {
  if { [cequal [get_selection] ""] } {
    show_status "No selection." steady
  } else {
    $Tgdb_cmd(print) [get_selection]
  }
}
button .f4.printc -bitmap @$bitmap_path/printc.xbm -command {
  if { [cequal [get_selection] ""] } {
    show_status "No selection." steady
  } else {
    $Tgdb_cmd(print) "*([get_selection])"
  }
}
button .f4.printx -bitmap @$bitmap_path/printx.xbm -command {
  $Tgdb_cmd(print) "/x"
}
button .f4.watch -bitmap @$bitmap_path/watch.xbm -command {
  if { [cequal [get_selection] ""] } {
    show_status "No selection." steady
  } else {
    $Tgdb_cmd(display) "[get_selection]"
  }
}
button .f4.finish -bitmap @$bitmap_path/finish.xbm \
  -command { $Tgdb_cmd(finish) }
button .f4.cont -bitmap @$bitmap_path/cont.xbm \
  -command { $Tgdb_cmd(continue) }
button .f4.next -bitmap @$bitmap_path/next.xbm \
  -command { $Tgdb_cmd(next) }
button .f4.step -bitmap @$bitmap_path/step.xbm \
  -command { $Tgdb_cmd(step) }
button .f4.break -bitmap @$bitmap_path/break.xbm \
  -command {
  if { [cequal [get_selection] ""] } {
    show_status "No selection." steady
  } else {
    $Tgdb_cmd(break) [get_selection]
  }
}
if { "[info commands blt_busy]" == "blt_busy" } {
  button .f4.stop -bitmap @$bitmap_path/stop.xbm -command {
    if { ([blt_busy isbusy] != "") && ![winfo exists .yesno] } {
      if { $use_expect } {
        if { [info exists tcl_platform] \
	     && ("$tcl_platform(platform)" == "unix") } {
	  exp_send ""
	} else {
	  kill SIGINT $gdb_pid
	}
      } elseif { $use_socket } {
        puts $gdb_desc \03
      } elseif { $use_remote } {
	kill SIGINT $gdb_pid
      } else {
	gdb_stop
      }
      set ProgCancelled 1
    } else {
      show_status "The application is currently not running." steady
    }
  }
} else {
  button .f4.stop -bitmap @$bitmap_path/stop.xbm -command {
    if { [info exists TkBusy] && ![winfo exists .yesno] } {
      if { $use_expect } {
        if { [info exists tcl_platform] \
	     && ("$tcl_platform(platform)" == "unix") } {
	  exp_send ""
	} else {
	  kill SIGINT $gdb_pid
	}
      } elseif { $use_socket } {
        puts $gdb_desc \03
      } elseif { $use_remote } {
	kill SIGINT $gdb_pid
      } else {
	gdb_stop
      }
      set ProgCancelled 1
    } else {
      show_status "The application is currently not running." steady
    }
  }
}
foreach but [winfo children .f4] {
  $but configure -activeforeground [lindex [$but configure -background] 4]
  $but configure -activebackground [lindex [$but configure -foreground] 4]
}
unset but
################################################################################
frame .f5 -relief flat
text .f5.text -exportselection 1 -yscrollcommand {.f5.scroll set} \
  -cursor $top_left_arrow
scrollbar .f5.scroll -orient vertical -command {.f5.text yview}
if { $colormodel == "color" } {
  .f5.text tag configure gdb_in -foreground black -background cyan
} else {
  .f5.text tag configure gdb_in -foreground white -background black
}
.f5.text tag lower gdb_in sel

################################################################################
#
# display widgets
#
################################################################################

pack append . .f0 {top fillx}
pack append .f0.f0 \
  .f0.f0.file left \
  .f0.f0.options left \
  .f0.f0.running left \
  .f0.f0.windows left \
  .f0.f0.cmds left \
  .f0.f0.thrds left \
  .f0.f0.help right
# .f0.f0.info left
pack .f0.f0 -side top -fill x
pack .title.title -side left -fill x -expand 1 -anchor center
pack .title.f1 -side left -fill both
pack .title.f1.line -side left -fill y
pack .title.f1.lentry -side left -fill y
pack .title -side top -fill x
pack .f2.lbl1 -side left
pack .f2.f.up -side top -ipadx 1 -ipady 1
pack .f2.f.dn -side bottom -ipadx 1 -ipady 1
pack .f2.f -side left
pack .f2.lbl2 -side left
pack .f2.search -side left -fill x -expand 1 -anchor w
pack .f2 -side top -fill x
pack .f3.text -side left -fill both -expand 1
pack .f3.scroll -side left -fill y
pack .f3 -side top -fill both -expand 1
pack .f1.status -side left -fill x -expand 1 -anchor w
pack .f1 -side top -fill x
pack .f4.list -side left -ipadx 1 -ipady 1
pack .f4.list_l -side left -ipadx 1 -ipady 1
pack .f4.list_d -side left -ipadx 1 -ipady 1
pack .f4.list_f -side left -ipadx 1 -ipady 1
pack .f4.frame_dn -side left -ipadx 1 -ipady 1 ;# -fill x -expand 1
pack .f4.frame_0 -side left -ipadx 1 -ipady 1 ;# -fill x -expand 1
pack .f4.frame_up -side left -ipadx 1 -ipady 1 ;# -fill x -expand 1
pack .f4.frame_in -side left -ipadx 1 -ipady 1
pack .f4.whatis -side left -ipadx 1 -ipady 1
pack .f4.print -side left -ipadx 1 -ipady 1
pack .f4.printc -side left -ipadx 1 -ipady 1
pack .f4.printx -side left -ipadx 1 -ipady 1
pack .f4.watch -side left -ipadx 1 -ipady 1
pack .f4.stop -side right -expand 1 -ipadx 1 -ipady 1
pack .f4.break -side right -ipadx 1 -ipady 1
pack .f4.finish -side right -ipadx 1 -ipady 1
pack .f4.step -side right -ipadx 1 -ipady 1
pack .f4.next -side right -ipadx 1 -ipady 1
pack .f4.cont -side right -ipadx 1 -ipady 1
pack .f4 -side top -fill x
pack .f5.text -side left -fill both -expand 1
pack .f5.scroll -side left -fill y
pack .f5 -side top -fill both

################################################################################
#
# enable menu traversal
#
################################################################################

if { $tk_version < 4.0 } {
  tk_menuBar \
    .f0.f0 \
      .f0.f0.file \
      .f0.f0.options \
      .f0.f0.running \
      .f0.f0.windows \
      .f0.f0.cmds \
      .f0.f0.thrds \
      .f0.f0.help
      #.f0.f0.info
}

################################################################################
#
# handle bindings
#
################################################################################

bind all <Control-c> {exit_tgdb}
bind all <Alt-Any-Key-Escape> {.f0.f0.options.m invoke $RaiseLabel}
bind all <Alt-Any-Key-a> {create_assembly_window}
bind all <Alt-Any-Key-c> {toggle_windows}
bind all <Alt-Any-Key-d> {create_memory_window}
bind all <Alt-Any-Key-h> {create_tgdb_help_window}
bind all <Alt-Any-Key-i> {help_about}
bind all <Alt-Any-Key-o> {create_pxmon_window}
bind all <Alt-Any-Key-r> {create_cpu_window}
bind all <Alt-Any-Key-s> {create_stack_window}
bind all <Alt-Any-Key-t> {raise_main_window}
bind all <Alt-Any-Key-w> {create_disp_window}
bind all <Alt-Any-Key-x> {create_shell_window}
bind all <Alt-Any-Key-z> {$Tgdb_cmd(run)}
if { $use_dbgwin } {
  bind all <Alt-Any-Key-p> {raise_debug_window}
}

#bind .f3.text <Alt-Any-Key-Escape> {.f0.f0.options.m invoke $RaiseLabel}
#bind .f3.text <Alt-Any-Key-a> {create_assembly_window}
#bind .f3.text <Alt-Any-Key-c> {toggle_windows}
#bind .f3.text <Alt-Any-Key-d> {create_memory_window}
#bind .f3.text <Alt-Any-Key-h> {create_tgdb_help_window}
#bind .f3.text <Alt-Any-Key-i> {help_about}
#bind .f3.text <Alt-Any-Key-o> {create_pxmon_window}
#bind .f3.text <Alt-Any-Key-r> {create_cpu_window}
#bind .f3.text <Alt-Any-Key-s> {create_stack_window}
#bind .f3.text <Alt-Any-Key-t> {raise_main_window}
#bind .f3.text <Alt-Any-Key-w> {create_disp_window}
#bind .f3.text <Alt-Any-Key-x> {create_shell_window}
#bind .f3.text <Alt-Any-Key-z> {$Tgdb_cmd(run)}
#if { $use_dbgwin } {
#  bind .f3.text <Alt-Any-Key-p> {raise_debug_window}
#}
bind .f3.text <Any-Key> { }
bind .f3.text <Any-Mod2-Key> { }
bind .f3.text <Control-Key-v> { }
bind .f3.text <Control-Key-d> { }
bind .f3.text <Control-Key-h> { }
bind .f3.text <BackSpace> { }
bind .f3.text <Control-c> {exit_tgdb}
bind .f3.text <Motion> { set Line [int [%W index @%x,%y]] }
bind .f3.text <2> [bind Text <2>]
bind .f3.text <B2-Motion> "[bind Text <B2-Motion>]"
bind .f3.text <B2-Motion> "+[bind .f3.text <Motion>]"
bind .f3.text <Double-2> {.f4.print invoke}
bind .f3.text <Up> {text_review [expr $Line - 2]}
bind .f3.text <Down> {text_review $Line}
bind .f3.text <Left> {
  text_review \
    [expr $Line - [lindex [split [wm geometry [winfo toplevel %W]] "x+"] 1]]
}
bind .f3.text <Right> {
  text_review \
    [expr $Line -2 + [lindex [split [wm geometry [winfo toplevel %W]] "x+"] 1]]
}
bind .f3.text <Control-Up> [bind .f3.text <Left>]
bind .f3.text <Control-Down> [bind .f3.text <Right>]
bind .f3.text <Return> [bind .f3.text <Down>]
bind .f3.text <KP_Enter> [bind .f3.text <Return>]
bind .f3.text <Key-j> [bind .f3.text <Down>]
bind .f3.text <Key-k> [bind .f3.text <Up>]
bind .f3.text <Key-h> [bind .f3.text <Left>]
bind .f3.text <Key-l> [bind .f3.text <Right>]
bind .f3.text <Prior> [bind .f3.text <Left>]
bind .f3.text <Next> [bind .f3.text <Right>]
bind .f3.text <Insert> { }
bind .f3.text <Delete> { }
bind .f3.text <Home> {text_review 0}
bind .f3.text <End> {text_review 10000000} ;# ugly, but fast...
bind .f3.text <3> {toggle_brk %W %x %y}
bind .f3.text <Control-1> {
  if { [expr2bpno [file tail $ThisFile]:[int [%W index @%x,%y]]] == "" } {
    if { $Cur_thread == 0 } {
      if { [$Tgdb_cmd(tbreak) [file tail $ThisFile]:[int [%W index @%x,%y]]] } {
        $Tgdb_cmd(continue)
      }
    } else {
      if { [$Tgdb_cmd(tbreak) \
             [file tail $ThisFile]:[int [%W index @%x,%y]] thr $Cur_thread] } {
        $Tgdb_cmd(continue)
      }
    }
  } else {
    $Tgdb_cmd(continue)
  }
}

################################################################################
# this binding realizes a popup menu within the text widget to set/goto marks
################################################################################

bind .f3.text <Shift-2> { popup_mark %X %Y }

proc popup_mark { x y } {
  global Popup Mark Line old_focus

  set Popup 0
  set old_focus(popup0) [focus]
  focus .none
  update
  toplevel .popup0 -cursor arrow
  bind .popup0 <Visibility> {focus .popup0; grab -global .popup0}
  wm geometry .popup0 "+$x+$y"
  wm transient .popup0 .
  wm focusmodel .popup0 active
  wm overrideredirect .popup0 1
  label .popup0.lbl -text {L: goto mark / R: set mark} -relief raised -border 2
  pack .popup0.lbl -side top -fill x
  foreach but "1 2 3 4 5" {
    if { ![info exists Mark($but)] } {
      set text "unset"
    } else {
      set text [string trim \
	[.f3.text get $Mark($but).0 "$Mark($but).0 lineend"]]
      if { [string length $text] > 30 } {
	set text "[string range $text 0 29]..."
      }
      set text "($Mark($but)) $text"
    }
    button .popup0.b$but -anchor w -text "#$but: $text"
    pack .popup0.b$but -side top -fill x
    bind .popup0.b$but <Shift-3> "set Mark($but) $Line; set Popup 1"
    if { $text != "unset" } {
      bind .popup0.b$but <Shift-1> "text_review [expr $Mark($but) - 1]; \
				   set Popup 1"
    }
  }
  tkwait visibility .popup0
  grab -global .popup0
  bind .popup0 <KeyRelease-Shift_L> {set Popup 1}
  bind .popup0 <KeyRelease-Shift_R> {set Popup 1}
  bind .popup0 <B2-KeyRelease-Shift_L> {set Popup 1}
  bind .popup0 <B2-KeyRelease-Shift_R> {set Popup 1}
  focus .popup0
  tkwait variable Popup
  destroy .popup0
  catch {focus $old_focus(popup0)}
}

################################################################################
# bind Ctrl-1..5 and Alt-1..5 as hotkeys for set/goto mark
################################################################################

set AltMark(1) 1
set AltMark(twosuperior) 2
set AltMark(threesuperior) 3
set AltMark(4) 4
set AltMark(5) 5
foreach mark "1 twosuperior threesuperior 4 5" {
  bind .f3.text <Control-Key-$AltMark($mark)> {
    show_status "Mark #%K set to line $Line."
    set Mark(%K) $Line
  }
  bind .f3.text <Alt-Any-Key-$mark> {
    if { [catch {text_review [expr $Mark($AltMark(%K)) - 1]}] } {
      show_status "Mark #$AltMark(%K) not set."
    } else {
      show_status "Goto mark #$AltMark(%K)."
    }
  }
}

################################################################################
# these bindings allow to select some consecutive characters to be used as
# a search string by using Shift-B3-Motion mouse movement
################################################################################

bind .f3.text <Shift-3> {
  if { [%W get @%x,%y] != "\n" } {
    %W tag add stag @%x,%y
    %W tag raise stag
    if { $colormodel == "color" } {
      %W tag configure stag -background green
    } else {
      %W tag configure stag -foreground white -background black
    }
    %W mark set smark @%x,%y
  }
}
bind .f3.text <Shift-B3-Motion> {
  tag_SearchString %W %x %y
}
bind .f3.text <Shift-ButtonRelease-3> {
  catch {set SearchString [%W get stag.first stag.last]}
  %W tag delete stag
  %W mark unset smark
}
bind .f3.text <Shift-B3-KeyRelease-Shift_L> \
  [bind .f3.text <Shift-ButtonRelease-3>]
bind .f3.text <Shift-B3-KeyRelease-Shift_R> \
  [bind .f3.text <Shift-ButtonRelease-3>]
bind .f3.text <Key-slash> { focus .f2.search }
bind .f3.text <Key-n> { search_string down }
bind .f3.text <Shift-N> { search_string up }

################################################################################
# these bindings handle text selection
################################################################################

bind .f3.text <1> {
  after 300 {bind %W <B1-Motion> {select_motion %%W %%x %%y}}
  bind %W <B1-Motion> { }
  select_c_expr %W %x %y
}
bind .f3.text <Double-1> {
  after 300 {bind %W <B1-Motion> {select_motion %%W %%x %%y}}
  bind %W <B1-Motion> { }
  select_c_bracket_expr %W %x %y
}
bind .f3.text <Triple-1> {
  after 300 {bind %W <B1-Motion> {select_motion %%W %%x %%y}}
  bind %W <B1-Motion> { }
  select_c_bracket_expr %W %x %y
  select_c_bracket_expr %W %x %y
}
bind .f3.text <B1-Motion> {select_motion %W %x %y}

################################################################################
#
# Bind lentry to Entry class default bindings, then override Any-KeyPress;
# this is important, since Any-KeyPress is more specific for lentry than
# for Entry (i.e. Entry bindings won't work anymore...)
#
################################################################################

foreach defbind [bind Entry] {
  bind .title.f1.lentry $defbind "[bind Entry $defbind]"
}
bind .title.f1.lentry <Any-KeyPress> {
  if { [string match \[0-9\] "%A"] } {
    %W insert insert %A
    tk_entrySeeCaret %W
  }
}
bind .title.f1.lentry <Escape> { set Line "" }
bind .title.f1.lentry <Return> adjust_curline
bind .title.f1.lentry <KP_Enter> [bind .title.f1.lentry <Return>]
#bind .title.f1.lentry <Any-Enter> {
# if { [focus] == ".none" || [winfo toplevel [focus]] == [winfo toplevel %W] } {
#   set old_focus(%W) [focus]
#   focus %W
# }
#}
bind .title.f1.lentry <Any-Enter> {
  if { [focus] != ".none" && [winfo toplevel [focus]] == [winfo toplevel %W] } {
    set old_focus(%W) [focus]
    focus %W
  }
}
#bind .title.f1.lentry <Any-Leave> {
# if { [focus] == ".none" || [winfo toplevel [focus]] == [winfo toplevel %W] } {
#   adjust_curline
#   catch {focus $old_focus(%W)}
# }
#}
bind .title.f1.lentry <Any-Leave> {
  if { [cequal [focus] %W] } {
    adjust_curline
    catch {focus $old_focus(%W)}
  }
}
bind .title.f1.lentry <Right> {
  %W icursor [expr [%W index insert]+1]; tk_entrySeeCaret %W
}
bind .title.f1.lentry <Left> {
  %W icursor [expr [%W index insert]-1]; tk_entrySeeCaret %W
}
bind .title.f1.lentry <Up> {%W icursor 0; %W view 0}
bind .title.f1.lentry <Down> {%W icursor end; tk_entrySeeCaret %W}
bindtags .title.f1.lentry .title.f1.lentry

################################################################################
#
# bind search entry
#
################################################################################

foreach defbind [bind Entry] {
  if { [lsearch -exact \
         "<Key-F10> <Key-BackSpace> <Key-Delete> <Key> \
	  <B2-Motion> <Button-2> <Button-1>" \
	 $defbind] >= 0 } {
    bind .f2.search $defbind "[bind Entry $defbind]"
  } else {
    bind .f2.search $defbind { }
  }
}
bind .f2.search <Any-Mod2-Key> [bind .f2.search <Any-Key>]
bind .f2.search <Escape> { set SearchString "" }
bind .f2.search <Return> { search_string down }
bind .f2.search <KP_Enter> [bind .f2.search <Return>]
#bind .f2.search <Any-Enter> {
# if { [focus] == ".none" || [winfo toplevel [focus]] == [winfo toplevel %W] } {
#   set old_focus(%W) [focus]
#   focus %W
# }
#}
bind .f2.search <Any-Enter> {
  if { [focus] != ".none" && [winfo toplevel [focus]] == [winfo toplevel %W] } {
    set old_focus(%W) [focus]
    focus %W
  }
}
#bind .f2.search <Any-Leave> {
# if { [focus] == ".none" || [winfo toplevel [focus]] == [winfo toplevel %W] } {
#   catch {focus $old_focus(%W)}
# }
#}
bind .f2.search <Any-Leave> {
  if { [cequal [focus] %W] } {
    catch {focus $old_focus(%W)}
  }
}
bind .f2.search <Right> {
  %W icursor [expr [%W index insert]+1]; tk_entrySeeCaret %W
}
bind .f2.search <Left> {
  %W icursor [expr [%W index insert]-1]; tk_entrySeeCaret %W
}
bind .f2.search <Up> {%W icursor 0; %W view 0}
bind .f2.search <Down> {%W icursor end; tk_entrySeeCaret %W}
bind .f2.search <3> {+
  if { [cequal [get_selection] ""] } {
    bell
    show_status "No selection."
  } else {
    set SearchString [get_selection]
  }
}
bindtags .f2.search .f2.search

################################################################################
#
# bind status entry
#
################################################################################

foreach bind [bind Entry] {
  if { [lsearch -exact "<Button-2> <B2-Motion> <Key-F10>" $bind] == -1 } {
    bind .f1.status $bind { }
  }
}
bind .f1.status <1> {split_srcwin_init %W %x %y}
bind .f1.status <B1-Motion> {split_srcwin_move %W %x %y}
bind .f1.status <ButtonRelease-1> {split_srcwin_end %W %x %y}

################################################################################
#
# bindings for help messages on buttons
#
################################################################################

foreach but [winfo children .f4] {
  bind $but <Any-Enter> [bind Button <Any-Enter>]
  bind $but <Any-Leave> [bind Button <Any-Leave>]
  bind $but <Any-Leave> {+
    if { !$Tgdb_busy && !$FreezeStatus} {
      show_status "" steady
    }
  }
}
bind .f4.list <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "List selected function." steady}}
bind .f4.list_l <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "List file that was displayed before the current one." steady}}
bind .f4.list_d <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "List disassembly of the line containing the selection." steady}}
bind .f4.list_f <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "List selected frame." steady}}
bind .f4.frame_dn <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "Select frame called by this one (go down one frame)." steady}}
bind .f4.frame_0 <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "Select innermost (bottom) frame." steady}}
bind .f4.frame_up <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "Select frame that called this one (go up one frame)." steady}}
bind .f4.frame_in <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "Show info about the selected frame." steady}}
bind .f4.whatis <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "Print data type of selected expression." steady}}
bind .f4.print <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "Print selected expression." steady}}
bind .f4.printc <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "Print selected pointer expression." steady}}
bind .f4.printx <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "Print last printed expression in hexadecimal format." steady}}
bind .f4.watch <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "Add selected expression to display list." steady}}
bind .f4.finish <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "Continue until selected frame returns (finishes)." steady}}
bind .f4.cont <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "Continue program execution." steady}}
bind .f4.next <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "Continue until next source line, skipping function calls." \
    steady}}
bind .f4.step <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "Continue until next source line, stepping into functions." \
    steady}}
bind .f4.break <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "Set breakpoint at selected function." steady}}
bind .f4.stop <Any-Enter> {+if { !$Tgdb_busy && !$FreezeStatus } {
  show_status "Stop program execution." steady}}

################################################################################
#
# bindings for .f5.text (gdb window)
#
################################################################################

# first, copy all class bindings
foreach bind [bind Text] {
  bind .f5.text $bind [bind Text $bind]
}
# then overload them as necessary
bind .f5.text <Any-Key> {
  if { ("%A" != "") && ("%A" != "{}") } {
    %W insert insert %A
    %W tag add gdb_in "insert -1c" insert
    %W yview -pickplace insert
  }
}
#bind .f5.text <Alt-Any-Key-Escape> {.f0.f0.options.m invoke $RaiseLabel}
bind .f5.text <Alt-Any-Key> [bind .f5.text <Any-Key>]
bind .f5.text <Alt-Any-Key-a> {create_assembly_window; break}
bind .f5.text <Alt-Any-Key-c> {toggle_windows; break}
bind .f5.text <Alt-Any-Key-d> {create_memory_window; break}
bind .f5.text <Alt-Any-Key-h> {create_tgdb_help_window; break}
bind .f5.text <Alt-Any-Key-i> {help_about; break}
bind .f5.text <Alt-Any-Key-o> {create_pxmon_window; break}
bind .f5.text <Alt-Any-Key-r> {create_cpu_window; break}
bind .f5.text <Alt-Any-Key-s> {create_stack_window; break}
bind .f5.text <Alt-Any-Key-t> {raise_main_window; break}
bind .f5.text <Alt-Any-Key-w> {create_disp_window; break}
bind .f5.text <Alt-Any-Key-x> {create_shell_window; break}
bind .f5.text <Alt-Any-Key-z> {$Tgdb_cmd(run); break}
if { $use_dbgwin } {
  bind .f5.text <Alt-Any-Key-p> {raise_debug_window; break}
}
bind .f5.text <BackSpace> {
  if { [lsearch -exact [%W tag names "insert -1c"] gdb_in] == -1 } {
    bell
  } else {
    tk_textBackspace %W
    %W yview -pickplace insert
  }
}
bind .f5.text <Delete> [bind .f5.text <BackSpace>]
bind .f5.text <Return> {
  if { [incr CmdInProgress] == 2 } {
    incr CmdInProgress -1
    break
  }
  set gdb_command [%W get "insert linestart" "insert lineend"]
  if { [set gdb_curpos [string first $prompt $gdb_command]] >= 0 } {
    incr gdb_curpos [string length $prompt]
    set gdb_command [string range $gdb_command $gdb_curpos end]
  }
  %W mark set insert "insert lineend"
  %W insert insert "\n"
  %W tag remove gdb_in insert-1c
  %W yview -pickplace insert
  if { [cequal $gdb_command ""] } {
    set gdb_history_last [expr ($gdb_history_nr - 1) %% $gdb_history_max]
    if { [info exists gdb_history($gdb_history_last)] } {
      set gdb_command $gdb_history($gdb_history_last)
      set xgdb_command $gdb_command
      expand_cmd xgdb_command xgdb_args
      if { [info exists gdb_dont_repeat($xgdb_command)] } {
	set gdb_command "echo"
      } elseif { [cequal $xgdb_command "x"] } {
	# don't repeat address args for last x command...
	set gdb_command "x ?noarg?"
      }
    } else {
      set gdb_command "echo"
    }
  } else {
    set gdb_history_last [expr ($gdb_history_nr - 1) %% $gdb_history_max]
    if {   ![info exists gdb_history($gdb_history_last)]
	|| ($gdb_command != $gdb_history($gdb_history_last)) } {
      set gdb_history($gdb_history_nr) $gdb_command
      set gdb_history_nr [expr ($gdb_history_nr + 1) %% $gdb_history_max]
    }
  }
  set gdb_history_spot $gdb_history_nr
  set gdb_history($gdb_history_nr) ""
  expand_cmd gdb_command gdb_args
  filter_output $gdb_command $gdb_args
  if {   [info exists gdb_class(user)]
      && ([lmatch -exact $gdb_class(user) $gdb_command] != "") } {
    update_windows new
  }
  incr CmdInProgress -1
}
bind .f5.text <KP_Enter> [bind .f5.text <Return>]
bind .f5.text <Up> {
  set gdb_history_last [expr ($gdb_history_spot - 1) %% $gdb_history_max]
  if { ![info exists gdb_history($gdb_history_last)] } {
    bell
  } else {
    if { $gdb_history_last == $gdb_history_nr } {
      bell
    } else {
      set gdb_history_spot $gdb_history_last
      set gdb_tmp [%W get "insert linestart" "insert lineend"]
      if { [set gdb_curpos [string first $prompt $gdb_tmp]] >= 0 } {
        incr gdb_curpos [string length $prompt]
        %W mark set insert [%W index "insert linestart + $gdb_curpos c"]
        %W delete insert "insert lineend"
        %W insert insert $gdb_history($gdb_history_spot)
        %W tag add gdb_in "insert linestart + $gdb_curpos c" insert
      } else {
        %W delete "insert linestart" "insert lineend"
        %W insert insert $gdb_history($gdb_history_spot)
        %W tag add gdb_in "insert linestart" insert
      }
      %W yview -pickplace insert
    }
  }
}
bind .f5.text <Down> {
  if { $gdb_history_spot == $gdb_history_nr } {
    bell
  } else {
    set gdb_history_last [expr ($gdb_history_spot + 1) %% $gdb_history_max]
    if { ![info exists gdb_history($gdb_history_last)] } {
      bell
    } else {
      set gdb_history_spot $gdb_history_last
      set gdb_tmp [%W get "insert linestart" "insert lineend"]
      if { [set gdb_curpos [string first $prompt $gdb_tmp]] >= 0 } {
        incr gdb_curpos [string length $prompt]
        %W mark set insert [%W index "insert linestart + $gdb_curpos c"]
        %W delete insert "insert lineend"
        %W insert insert $gdb_history($gdb_history_spot)
        %W tag add gdb_in "insert linestart + $gdb_curpos c" insert
      } else {
        %W delete "insert linestart" "insert lineend"
        %W insert insert $gdb_history($gdb_history_spot)
        %W tag add gdb_in "insert linestart" insert
      }
      %W yview -pickplace insert
    }
  }
}
bind .f5.text <Control-a> {
  set gdb_tmp [%W get "insert linestart" "insert lineend"]
  if { [set gdb_curpos [string first $prompt $gdb_tmp]] >= 0 } {
    incr gdb_curpos [string length $prompt]
    %W mark set insert [%W index "insert linestart + $gdb_curpos c"]
  } else {
    %W mark set insert [%W index "insert linestart"]
  }
  %W yview -pickplace insert
}
bind .f5.text <Control-b> {
  if { [lsearch -exact [%W tag names "insert -1c"] gdb_in] == -1 } {
    bell
  } else {
    %W mark set insert "insert-1c"
    %W yview -pickplace insert
  }
}
bind .f5.text <Control-c> {
  set gdb_tmp [%W get "insert linestart" "insert lineend"]
  if { [set gdb_curpos [string first $prompt $gdb_tmp]] >= 0 } {
    incr gdb_curpos [string length $prompt]
    set gdb_tmp [string range $gdb_tmp $gdb_curpos end]
  }
  if { [cequal $gdb_tmp ""] } {
    exit_tgdb
  } else {
    %W mark set insert "insert lineend"
    %W insert insert "\n"
    %W tag remove gdb_in "insert-1c"
    %W insert end $prompt
    %W yview -pickplace end
    set gdb_history_spot $gdb_history_nr
  }
  break
}
bind .f5.text <Control-d> {
  if { [lsearch -exact [%W tag names insert] gdb_in] == -1 } {
    bell
  } else {
    %W delete insert
    %W yview -pickplace insert
  }
}
bind .f5.text <Control-e> {
  %W mark set insert "insert lineend"
}
bind .f5.text <Control-f> {
  if { [lsearch -exact [%W tag names insert] gdb_in] == -1 } {
    bell
  } else {
    %W mark set insert "insert+1c"
    %W yview -pickplace insert
  }
}
bind .f5.text <Control-g> {bell}
bind .f5.text <Control-h> [bind .f5.text <BackSpace>]
bind .f5.text <Control-j> [bind .f5.text <Return>]
bind .f5.text <Control-k> {
  %W delete insert "insert lineend"
  %W yview -pickplace insert
}
bind .f5.text <Control-l> {
  if { $Tgdb_option(gdbRO) } {
    %W yview [int [%W index "insert-1l"]].0
  } else {
    %W delete 1.0 "insert linestart"
    %W yview 1.0
  }
}
bind .f5.text <Control-m> [bind .f5.text <Return>]
bind .f5.text <Control-n> [bind .f5.text <Down>]
bind .f5.text <Control-o> { }
bind .f5.text <Control-p> [bind .f5.text <Up>]
bind .f5.text <Control-q> { }
bind .f5.text <Control-r> {focus .none; reverse_i_search}
bind .f5.text <Control-s> { }
bind .f5.text <Control-t> { }
bind .f5.text <Control-u> { }
bind .f5.text <Control-v> { }
bind .f5.text <Control-w> { }
bind .f5.text <Control-x> { }
bind .f5.text <Control-y> { }
bind .f5.text <Control-z> { }
bind .f5.text <Left> [bind .f5.text <Control-b>]
bind .f5.text <Right> [bind .f5.text <Control-f>]
bind .f5.text <Shift-Left> [bind .f5.text <Control-a>]
bind .f5.text <Shift-Right> [bind .f5.text <Control-e>]
bind .f5.text <End> [bind .f5.text <Control-e>]
bind .f5.text <Home> [bind .f5.text <Control-a>]
#bind .f5.text <Any-Enter> {
# if { [focus] == ".none" || [winfo toplevel [focus]] == [winfo toplevel %W] } {
#   set old_focus(f5.text) [focus]
#   focus %W
# }
#}
bind .f5.text <Any-Enter> {
  if { [focus] != ".none" && [winfo toplevel [focus]] == [winfo toplevel %W] } {
    set old_focus(%W) [focus]
    focus %W
  }
}
#bind .f5.text <Any-Leave> {
# if { [focus] == ".none" || [winfo toplevel [focus]] == [winfo toplevel %W] } {
#   catch {focus $old_focus(f5.text)}
# }
#}
bind .f5.text <Any-Leave> {
  if { [cequal [focus] %W] } {
    catch {focus $old_focus(%W)}
  }
}
bind .f5.text <Prior> {
  %W yview [expr [int [%W index @0,0]] - \
    ([winfo height %W] - [lindex [%W conf -bd] 4]) / $WinSize(.,y2)].0
}
bind .f5.text <Next> {
  %W yview [expr [int [%W index @0,0]] -2 + \
    ([winfo height %W] - [lindex [%W conf -bd] 4]) / $WinSize(.,y2)].0
}
bind .f5.text <Shift-Prior> [bind .f5.text <Prior>]
bind .f5.text <Shift-Next> [bind .f5.text <Next>]
bind .f5.text <Shift-Up> [bind .f5.text <Prior>]
bind .f5.text <Shift-Down> [bind .f5.text <Next>]
bind .f5.text <Tab> {command_line_completion}

################################################################################
# these bindings handle text selection...
################################################################################

bind .f5.text <1> {
  after 300 {bind %W <B1-Motion> {select_motion %%W %%x %%y}}
  bind %W <B1-Motion> { }
  %W mark set anchor @%x,%y
  select_c_expr %W %x %y
}
bind .f5.text <Double-1> {
  after 300 {bind %W <B1-Motion> {select_motion %%W %%x %%y}}
  bind %W <B1-Motion> { }
  select_c_bracket_expr %W %x %y
}
bind .f5.text <Triple-1> {
  after 300 {bind %W <B1-Motion> {select_motion %%W %%x %%y}}
  bind %W <B1-Motion> { }
  select_c_bracket_expr %W %x %y
  select_c_bracket_expr %W %x %y
}
bind .f5.text <B1-Motion> {select_motion %W %x %y}
bind .f5.text <3> {insert_selection %W}
bind .f5.text <Double-2> {.f4.print invoke}

################################################################################
################################################################################
#
# Miscellaneous procedures
#
################################################################################
################################################################################

################################################################################
#
# procedures for selecting c expressions in text widgets
#
################################################################################

################################################################################
# select a simple c expression (by pressing button 1)
################################################################################

proc select_c_expr { w x y } {
  global SelAnchor

  set idx [$w index @$x,$y]
  set SelAnchor($w) $idx
  catch {selection clear .}
  set str [$w get "$idx linestart" "$idx lineend"]
  set delims " !%^&*()\$\[\]+-=~|;:{}\\,./#<>@?\"\n\t"
  regexp {(.*)\.(.*)} $idx {} line pos
  set len [string length $str]
  # find first position of c expression under the cursor
  for { set i $pos } { $i >= 0 } { incr i -1 } {
    if { [string match "*\\[cindex $str $i]*" $delims] } {
      break
    }
  }
  incr i; set beg [expr ($i > $pos) ? $pos : $i]
  # find last position of c expression under the cursor
  for { set i $pos } { $i < $len } { incr i } {
    if { [string match "*\\[cindex $str $i]*" $delims] } {
      break
    }
  }
  set end $i
  $w tag add sel $line.$beg $line.$end
}

################################################################################
# select a c expression containing brackets, parenthesis, '.' or '->' (Double-1)
# if we come here, something should have been selected due to the first
# mouse click - if not, we'll just return, otherwise we try to see if the
# first non-whitespace character after the last selected one is a bracket
# ("(" or "["). If this is the case, we try to select all characters up to
# the matching bracket or parenthesis; we handle strings as well as nested
# brackets and escape sequences. In case of '.' or '->' we select the
# following simple c expression.
################################################################################

proc select_c_bracket_expr { w x y } {
  if { [cequal [get_selection] ""] } {
    return
  }
  #
  # find next non-whitespace character first
  #
  set beg [lindex [$w tag ranges sel] 0]
  set end [lindex [$w tag ranges sel] 1]
  set ready 0
  while { !$ready } {
    set str [$w get "$end linestart" "$end lineend"]
    if { [set len [clength $str]] == 0 } {
      return
    }
    for { regexp {.*\.(.*)} $end {} idx } { $idx < $len } { incr idx } {
      if { ![cequal [string trim [cindex $str $idx]] ""] } {
        set ready 1
        break
      }
    }
    if { $ready } {
      set char [cindex $str $idx]
      if { $char == "(" } {
	set match ")"
      } elseif { $char == "\[" } {
	set match "\]"
      } elseif {   [cequal $char .]
	    || ([cequal $char -] && [cequal [cindex $str [incr idx]] >]) } {
	set ready 0
	while { !$ready } {
	  for { incr idx } { $idx < $len } { incr idx } {
	    if { ![cequal [string trim [cindex $str $idx]] ""] } {
	      set ready 1
	      break
	    }
	  }
	  if { $idx == $len } {
	    set idx 0
	    set end [expr [int $end] + 1].0
	    set str [$w get "$end linestart" "$end lineend"]
	    if { [set len [clength $str]] == 0 } {
	      return
	    }
	  }
	}
	set delims " !%^&*()\$\[\]+-=~|;:{}\\,./#<>@?\"\n\t"
	for { } { $idx < $len } { incr idx } {
	  if { [string match "*\\[cindex $str $idx]*" $delims] } {
	    break
	  }
	}
	catch {selection clear .}
	$w tag add sel $beg "$end linestart +$idx c"
	return
      } else {
        return
      }
    } else {
      set end [expr [int $end] + 1].0
    }
  }
  #
  # now find the matching bracket
  #
  set end [int $end].[incr idx]
  set nestlvl 1
  set in_str 0
  set ready 0
  while { !$ready } {
    set str [$w get "$end linestart" "$end lineend+1c"]
    if { [set len [string length $str]] == 0 } {
      return
    }
    for { regexp {.*\.(.*)} $end {} idx } { $idx < $len } { incr idx } {
      if { [cindex $str $idx] == "\\" } {
	incr idx 1
      } elseif { $in_str } {
        if { [cindex $str $idx] == "\"" } {
	  set in_str 0
	}
      } elseif { [cindex $str $idx] == "\"" } {
	set in_str 1
      } elseif { [cindex $str $idx] == "$char" } {
	incr nestlvl
      } elseif { [cindex $str $idx] == "$match" } {
	if { [incr nestlvl -1] == 0 } {
          set ready 1
          break
	}
      }
    }
    if { $ready } {
      set end [int $end].$idx
    } else {
      set end [expr [int $end] + 1].0
    }
  }
  catch {selection clear .}
  $w tag add sel $beg $end+1c
}

################################################################################
# select characters in the usual way by dragging the mouse with button 1 pressed
################################################################################

proc select_motion { w x y } {
  global SelAnchor

  catch {selection clear .}
  set beg $SelAnchor($w)
  set end [$w index @$x,$y]
  if { [$w compare $beg <= $end] } {
    $w tag add sel $beg $end+1c
  } else {
    $w tag add sel $end $beg+1c
  }
}

################################################################################
# return selected characters (or "" if nothing is selected - instead of griping)
################################################################################

proc get_selection {} {
  if { [catch {set selection [selection get]}] } {
    return ""
  }
  regsub -all "(\n|\t| )+" $selection " " selection
  return "$selection"
}

################################################################################
# insert selection (if there is one) to the given text widget
################################################################################

proc insert_selection { w } {
  if { ![cequal [set selection [get_selection]] ""] } {
    set beg [$w index end-1c]
    $w insert end $selection
    $w tag add gdb_in $beg end-1c
    $w yview -pickplace end
  } else {
    bell
    show_status "No selection."
  }
}

################################################################################
# list selected file
################################################################################

proc list_sel_file {} {
  global Tgdb_cmd

  if { [cequal [set file [get_selection]] ""] } {
    show_status "No selection." steady
  } else {
    $Tgdb_cmd(list) $file
  }
}

################################################################################
# list last file (the one displayed before the current one)
################################################################################

proc list_last_file {} {
  global LastFile LastLine

  if { [cequal $LastFile ""] } {
    show_status "No previously visited file." steady
  } else {
    set lastline $LastLine
    load_file $LastFile $LastLine
    .f3.text yview $lastline.0
  }
}


################################################################################
#
# procedures to keep track of the current line
#
################################################################################

################################################################################
# adjust current line in text widget
################################################################################

proc adjust_curline {{how -pickplace}} {
  global Line

  if { [catch {expr $Line}] } {
    set Line 1
  }
  if { $Line <= 0} {
    set Line 1
  } else {
    set end [expr [int [.f3.text index end-1c]] - 1]
    if { $Line > $end } {
      set Line $end
    }
  }
  set lastline [int [.f3.text index @0,[winfo height .f3.text]]]
  if { ($Line >= [int [.f3.text index @0,0]]) && ($Line <= $lastline) } return
  if { $how == "-pickplace" } {
    if { $Line > $lastline } {
      .f3.text yview -pickplace [expr $Line + 1].0
    } else {
      .f3.text yview -pickplace [expr $Line - 1].0
    }
  } else {
    if { $Line > $lastline } {
      .f3.text yview [expr $Line + 1].0
    } else {
      .f3.text yview [expr $Line - 1].0
    }
  }
}
################################################################################
proc text_review { ypos } {
  global Line

  set Line [expr $ypos + 1]
  adjust_curline
}
################################################################################
proc text_scroll { args } {
  global Line

  #set Line [expr $ypos + 1]
  #adjust_curline -nopickplace ;# don't use -pickplace for scrollbar commands!
  eval .f3.text yview $args
  set Line [int [.f3.text index @0,0]]
}

################################################################################
#
# display a message in the status line (and delete it after a while, if req'd)
#
################################################################################

proc show_status { msg {how_long 2000} } {
  global TextStatus StatusID

  regsub -all "\n" $msg " " msg
  catch {after cancel $StatusID}
  set TextStatus $msg
  if { "$how_long" != "steady" } {
    set StatusID [after $how_long {set TextStatus ""}]
  }
  update idletasks
}

################################################################################
#
# tag a selected range of characters for later use as a search string
#
################################################################################

proc tag_SearchString { window xpos ypos } {
  global colormodel

  if { [catch {set beg [$window index smark]}] } {
    set beg @$xpos,$ypos
    if { [$window get $beg] == "\n" } {
      return
    }
    $window tag add stag $beg
    $window tag raise stag
    if { $colormodel == "color" } {
      $window tag configure stag -background green
    } else {
      $window tag configure stag -foreground white -background black
    }
    $window mark set smark $beg
  }
  set end [lindex [$window tag ranges stag] 1]
  set cur "@$xpos,$ypos"
  if { [$window compare $cur < "$beg linestart"] } {
    set cur "$beg linestart"
  }
  if { [$window compare $cur >= "$beg lineend"] } {
    set cur "$beg lineend - 1 char"
  }
  if { [$window compare $cur < $beg] } {
    set end "$beg + 1 char"
    set beg "$cur"
  } else {
    set end "$cur + 1 char"
  }
  $window tag remove stag 1.0 $beg
  $window tag remove stag $end end
  $window tag add stag $beg $end
}

################################################################################
#
# reset (t)gdb's command history
#
################################################################################

proc clear_command_history {} {
  global gdb_history gdb_history_nr gdb_history_spot

  if { $gdb_history_nr == 0 } {
    show_status "The command line history is empty."
  } else {
    if { [yes_no_box "Clear command line history?"] } {
      unset gdb_history
      set gdb_history_nr 0
      set gdb_history_spot 0
      set gdb_history(0) ""
    }
  }
}

################################################################################
#
# if colormodel is mono, make the current line blink if it contains a
# breakpoint tag and the current PC
#
################################################################################

proc blink_line {} {
  global BlinkingPC BlinkingAsmPC

  if { $BlinkingPC } {
    set fg [lindex [.f3.text tag configure mytag -foreground] 4]
    set bg [lindex [.f3.text tag configure mytag -background] 4]
    .f3.text tag configure mytag -foreground $bg
    .f3.text tag configure mytag -background $fg
  }
  if { $BlinkingAsmPC } {
    if { ![winfo exists .asm] } {
      set BlinkingAsmPC 0
    } else {
      set fg [lindex [.asm.f1.txt tag configure mytag -foreground] 4]
      set bg [lindex [.asm.f1.txt tag configure mytag -background] 4]
      .asm.f1.txt tag configure mytag -foreground $bg
      .asm.f1.txt tag configure mytag -background $fg
    }
  }
  after 500 blink_line
}

################################################################################
#
# load a file and display it in the source text widget starting at line cur_line
# and mark this line if highlight isn't "no"
#
################################################################################

proc load_file { name cur_line {highlight no} } {
  global Contents Line ThisFile TextBindings LentryBindings SearchBindings
  global Mark Marks Tags SearchLastLine LastFile LastLine colormodel
  global BlinkingPC ModTime WinSize Tgdb_option
  global Cur_thread LastThread Ttags

  proc set_filemenu_state { state } {
    .f0.f0.file.m entryconfigure { Load program } -state $state
    .f0.f0.file.m entryconfigure { Load executable file } -state $state
    .f0.f0.file.m entryconfigure { Load symbol file } -state $state
    .f0.f0.file.m entryconfigure { Load core file } -state $state
    .f0.f0.file.m entryconfigure { Load gdb source file } -state $state
    .f0.f0.file.m entryconfigure { Load text file } -state $state
    .f0.f0.file.m entryconfigure { Visit } -state $state
    .f0.f0.file.m entryconfigure { Save gdb window } -state $state
    .f0.f0.file.m entryconfigure { Save gdb init file } -state $state
    .f0.f0.file.m entryconfigure { Add source directory } -state $state
    .f0.f0.file.m entryconfigure { Change working directory } -state $state
  }

  if { [cequal $name ""] } {
    if { ![cequal $ThisFile ""] } {
      show_status "File selection cancelled." 2000
    }
    return
  }
  if { ![file readable $name] } {
    set name [file2nfspath $name]
  }
  if { [cequal $name $ThisFile] && ([file mtime $name] > $ModTime($name)) } {
    # reload current file
    show_status "Reloading file \"$name\"..." steady
    update idletasks
    set_filemenu_state disabled
    unset Contents($name)
    if { [catch {set desc [open $name r]}] } {
      show_status "No such file \"$name\"." 4000
      set_filemenu_state normal
      return
    }
    if { [catch {set Contents($name) [read $desc]}] } {
      show_status "Error reading file \"$name\"." 4000
      close $desc
      catch {unset Contents($name)}
      set_filemenu_state normal
      return
    }
    close $desc
    if { $Tgdb_option(RemCRs) } {
      regsub -all "
\n" $Contents($name) "\n" Contents($name)
    }
    set ModTime($name) [file mtime $name]
    .f3.text configure -state normal
    .f3.text delete 1.0 end
    .f3.text insert 1.0 $Contents($name)
    .f3.text configure -state disabled
    if { $colormodel == "color" } {
      .f3.text tag configure brktag -background red
    } else {
      .f3.text tag configure brktag -foreground white -background black
    }
    .f3.text tag lower brktag
    set_filemenu_state normal
    update_bpts
    show_status ""
  }

  if { [cequal $name $ThisFile] } {
    # the following lines provide a better "pickplace" behaviour...
    set topline [int [.f3.text index @0,0]]
    set lstline [expr $topline - 1 \
      + (([winfo height .f3.text] - [lindex [.f3.text conf -bd] 4]) \
      / $WinSize(.,y))]
    if { ($cur_line <= $topline) && ($cur_line >= [expr $topline - 5]) } {
      set Line [expr $topline -8]; adjust_curline exact
    } elseif { ($cur_line >= $lstline) && ($cur_line <= [expr $lstline + 5]) } {
      set Line [expr $topline -8]; adjust_curline exact
    }
    set Line $cur_line; adjust_curline
    if { $Cur_thread != $LastThread } {
      catch {.f3.text tag delete tbrktag}
      if { [info exists Ttags($name,$Cur_thread)] } {
	if { $colormodel == "color" } {
	  .f3.text tag configure tbrktag -background pink
	} else {
	  .f3.text tag configure tbrktag -foreground white -background black
	}
	.f3.text tag lower tbrktag
        foreach tag $Ttags($name,$Cur_thread) {
	  .f3.text tag add tbrktag $tag "$tag+1l"
	}
      }
      set LastThread $Cur_thread
    }
    if { ($highlight != "no") && ($cur_line != 0) } {
      .f3.text tag delete mytag
      .f3.text tag add mytag $Line.0 "$Line.0 +1 line"
      .f3.text tag lower mytag sel
      .f3.text tag raise mytag brktag
      if { $colormodel == "color" } {
	if { [lsearch -regexp [.f3.text tag names $Line.0] ^t?brktag] == -1 } {
	  .f3.text tag configure mytag -background green
	} else {
	  .f3.text tag configure mytag -background cyan
	}
      } else {
	.f3.text tag configure mytag -foreground white -background black
	if { [lsearch -exact [.f3.text tag names $Line.0] brktag] == -1 } {
	  set BlinkingPC 0
	} else {
	  set BlinkingPC 1
	}
      }
    }
    return
  }

  # disable menu entries
  set_filemenu_state disabled

  # see if we should force reloading
  if {   [info exists Contents($name)]
      && ([file mtime $name] > $ModTime($name)) } {
    show_status "Reloading file \"$name\"..."
    update idletasks
    unset Contents($name)
    update_bpts
  }

  # only load the file if it isn't loaded yet
  if { ![info exists Contents($name)] } {
    if { [catch {set desc [open $name r]}] } {
      show_status "No such file \"$name\"." 4000
      set_filemenu_state normal
      return
    }
    if { [catch {set Contents($name) [read $desc]}] } {
      show_status "Error reading file \"$name\"." 4000
      close $desc
      catch {unset Contents($name)}
      set_filemenu_state normal
      return
    }
    close $desc
    if { $Tgdb_option(RemCRs) } {
      regsub -all "
\n" $Contents($name) "\n" Contents($name)
    }
    set ModTime($name) [file mtime $name]
  }

  # remember current file and line number
  set LastFile $ThisFile
  set LastLine [expr [int [.f3.text index @0,0]] - 1]
  set LastThread $Cur_thread

  # forget last string search position (if any)
  catch {unset SearchLastLine}

  # save old marks and selection tags
  if { ![cequal $ThisFile ""] } {
    foreach i "1 2 3 4 5" {
      catch {set Marks($ThisFile,$i) $Mark($i)}
    }
    catch {set Tags($ThisFile,sel) [.f3.text tag ranges sel]}
  }
  # remove all current marks and tags
  foreach i "1 2 3 4 5" { catch {unset Mark($i)} }
  catch {.f3.text tag delete [.f3.text tag names]}

  .f3.text configure -state normal
  .f3.text delete 1.0 end
  .f3.text insert 1.0 $Contents($name)
  .f3.text configure -state disabled
  if { $colormodel == "color" } {
    .f3.text tag configure brktag -background red
    .f3.text tag configure tbrktag -background pink
  } else {
    .f3.text tag configure brktag -foreground white -background black
    .f3.text tag configure tbrktag -foreground white -background black
  }
  .f3.text tag lower brktag
  .f3.text tag lower tbrktag

  # restore old marks and tags if available
  foreach i "1 2 3 4 5" { catch {set Mark($i) $Marks($name,$i)} }
  catch {
    set tags $Tags($name,sel)
    set len [llength $tags]
    for { set b 0; set e 1 } { $e < $len } { incr b 2; incr e 2 } {
      .f3.text tag add sel [lindex $tags $b] [lindex $tags $e]
    }
  }
  if { ![info exists Tags($name,brktag)] } {
    set Tags($name,brktag) ""
  } else {
    foreach tag $Tags($name,brktag) {
      .f3.text tag add brktag $tag $tag+1l
    }
  }
  if { ![info exists Ttags($name,$Cur_thread)] } {
    set Ttags($name,$Cur_thread) ""
  } else {
    foreach tag $Ttags($name,$Cur_thread) {
      .f3.text tag add tbrktag $tag $tag+1l
    }
  }

  set Line $cur_line; adjust_curline
  if { ($highlight != "no") && ($cur_line != 0) } {
    .f3.text tag delete mytag
    .f3.text tag add mytag $Line.0 "$Line.0 +1 line"
    .f3.text tag lower mytag sel
    .f3.text tag raise mytag brktag
    if { $colormodel == "color" } {
      if { [lsearch -regexp [.f3.text tag names $Line.0] ^t?brktag] == -1 } {
        .f3.text tag configure mytag -background green
      } else {
        .f3.text tag configure mytag -background cyan
      }
    } else {
      .f3.text tag configure mytag -foreground white -background black
      if { [lsearch -regexp [.f3.text tag names $Line.0] ^t?brktag] == -1 } {
	set BlinkingPC 0
      } else {
	set BlinkingPC 1
      }
    }
  }
  set ThisFile $name
  .title.f1.lentry configure -state normal

  # restore bindings for text/line/search entry widgets
  foreach bind [array names TextBindings] {
    bind .f3.text $bind $TextBindings($bind)
  }
  foreach bind [array names LentryBindings] {
    bind .title.f1.lentry $bind $LentryBindings($bind)
  }
  foreach bind [array names SearchBindings] {
    bind .f2.search $bind $SearchBindings($bind)
  }

  # generate new file list for Menu->Visit entry and re-enable entries
  catch {.f0.f0.file.m.visit delete 0 last}
  set pwd "[pwd]/"; set pwdlen [string length $pwd]
  foreach name [lsort [array names Contents]] {
    if { [string first $pwd $name] == 0 } {
      set fname [string range $name $pwdlen end]
    } else {
      set fname $name
    }
    if { [cequal [file tail $fname] LICENSE] } {
      set fname "LICENSE"
    }
    .f0.f0.file.m.visit add command -label "$fname" -command "load_file $name 0"
  }
  set_filemenu_state normal
}

################################################################################
#
# load file LICENSE and display a copyright/license dialog box
# Note: as the LICENSE file tells you, you are not allowed to change tgdb's
#       sources before you have paid the shareware fee!
#
################################################################################

proc show_copyright_and_license_policy {} {
  global tgdb_path std_out Contents Tags ModTime

  set name $tgdb_path/LICENSE
  if { ![file readable $name] } {
    puts $std_out "Cannot read tgdb's LICENSE file."
    exit 1
  }
  if { [catch {set desc [open $name r]}] } {
    puts $std_out "Cannot open tgdb's LICENSE file."
    exit 1
  }
  if { [catch {set Contents($name) [read $desc]}] } {
    puts $std_out "Cannot read tgdb's LICENSE file."
    exit 1
  }
  close $desc
  set Tags($name,brktag) ""
  set ModTime($name) [file mtime $name]

  catch {.f0.f0.file.m.visit delete 0 last}
  .f0.f0.file.m.visit add command -label LICENSE -command "load_file $name 0"
  .f0.f0.file.m entryconfigure { Visit } -state normal
  show_license_box
}

################################################################################
#
# search a string upwards/downwards from the current line
#
################################################################################

proc search_string { direction } {
  global Interrupt

  set Interrupt 0
  set cursor_before_search [lindex [.f3.text configure -cursor] 4]
  .f3.text configure -cursor watch
  set bind_before_search [bind .f3.text <Delete>]
  bind .f3.text <Delete> {set Interrupt 1}
  search_string2 $direction
  .f3.text configure -cursor $cursor_before_search
  bind .f3.text <Delete> $bind_before_search
  if { $Interrupt } {
    set Interrupt 0
    show_status "Interrupt."
  }
  unset bind_before_search cursor_before_search
}
################################################################################
proc search_string2 { direction } {
  global Line Contents ThisFile SearchString SearchLastLine SearchLastIdx
  global Interrupt colormodel


  if { [cequal $SearchString ""] } {
    show_status "No string to search for."
    return
  }
  if { [catch {regexp -indices -- $SearchString ""} status] } {
    show_status $status steady
    return
  } else {
    show_status "Searching..." steady
  }
  set bol_only [regexp {^\^} $SearchString]
  set pos [.f3.text index end-1c]
  set end [expr [string range $pos 0 [expr [string first "." $pos] - 1]]]
  set wrapped 0
  set lno $Line

  if { $direction == "down" } {
    if { [info exists SearchLastLine] && ($lno == $SearchLastLine) } {
      if { $bol_only } {
	incr lno
	set beg 0
      } else {
        set beg [expr $SearchLastIdx + 1]
	if { $beg >= [string length [.f3.text get $lno.0 "$lno.0 lineend"]] } {
	  incr lno
	  set beg 0
	}
      }
    } else {
      set beg 0
    }
    while {1} {
      if { [expr $lno % 20] == 0 } {
        update
        if { $Interrupt } return
      }
      if { $lno > $Line && $wrapped } {
        show_status "Pattern not found."
        return
      }
      if { $lno == $end } {
        set lno 1
	set beg 0
        set wrapped 1
      }
      set str [.f3.text get $lno.$beg "$lno.0 lineend"]
      if { [regexp -indices -- $SearchString $str idx] } {
        text_review [expr $lno - 1]
	#.f3.text yview -pickplace [expr $lno - 1]
	catch {.f3.text tag delete search$SearchLastLine$SearchLastIdx}
        set p0 [expr [lindex $idx 0] + $beg]
        set p1 [expr [lindex $idx 1] + $beg]
        set SearchLastLine $lno
        set SearchLastIdx $p0
        .f3.text tag add search$lno$p0 $lno.$p0 "$lno.$p1 + 1 char"
	if { $colormodel == "color" } {
          .f3.text tag configure search$lno$p0 -background yellow
	} else {
          .f3.text tag configure search$lno$p0 -foregroun white -backgroun black
	}
        .f3.text tag raise search$lno$p0
         after 4000 "catch \{.f3.text tag delete search$lno$p0\}"
	if { $wrapped } { show_status "Wrapped." 3000 } else { show_status "" }
        return
      }
      set beg 0
      incr lno
    }
  } elseif { $direction == "up" } {
    if { [info exists SearchLastLine] && ($lno == $SearchLastLine) } {
      if { $bol_only || ($SearchLastIdx == 0) } {
	incr lno -1
      }
    } 
    while {1} {
      if { [expr $lno % 20] == 0 } {
        update
        if { $Interrupt } return
      }
      if { ($lno < $Line) && $wrapped } {
        show_status "Pattern not found."
        return
      }
      if { $lno == 0 } {
        set lno $end 
        set wrapped 1
      }
      if { ($lno == $Line) && !$wrapped } {
        set str [.f3.text get $lno.0 "$lno.0 lineend"]
	if { [info exists SearchLastLine] && ($lno == $SearchLastLine) } {
	  set last $SearchLastIdx
        } else {
          set last [string length $str]
	  if { $last == 0 } { set last 1 } ;# so we can find ".*$" in ""
        }
      } else {
        set str [.f3.text get $lno.0 "$lno.0 lineend"]
        set last [string length $str]
	if { $last == 0 } { set last 1 } ;# so we can find ".*$" in ""
      }
      if { [regexp -indices -- $SearchString $str idx] } {
        set beg [lindex $idx 0]
	set old_idx $idx
	if { !$bol_only } {
	  incr last -1
          set str [string range $str [expr $beg + 1] end]
	  while {   ![cequal $str ""]
		 && [regexp -indices -- $SearchString $str idx] } {
	    if { [expr $beg + [lindex $idx 0]] >= $last } {
	      set idx $old_idx
	      break
	    }
	    set beg [expr $beg + [lindex $idx 0] + 1]
	    set str [string range $str [expr [lindex $idx 0] + 1] end]
	    set old_idx $idx
	  }
	  incr last
	  set idx $old_idx
	}
	if { $beg < $last } {
          text_review [expr $lno - 1]
	  #.f3.text yview -pickplace [expr $lno - 1]
	  catch {.f3.text tag delete search$SearchLastLine$SearchLastIdx}
          set p0 $beg
          set p1 [expr [lindex $idx 1] - [lindex $idx 0] + $beg]
          set SearchLastLine $lno
          set SearchLastIdx $p0
          .f3.text tag add search$lno$p0 $lno.$p0 "$lno.$p1 + 1 char"
	  if { $colormodel == "color" } {
            .f3.text tag configure search$lno$p0 -background yellow
	  } else {
            .f3.text tag configure search$lno$p0 -forgroun white -backgrou black
	  }
          .f3.text tag raise search$lno$p0
           after 4000 "catch \{.f3.text tag delete search$lno$p0\}"
	  if { $wrapped } { show_status "Wrapped." 3000} else { show_status "" }
          return
	}
      }
      incr lno -1
    }
  } else {
    show_status "Internal error: wrong search direction \"$direction\"." steady
  }
}

################################################################################
# display the file select box to get a file for use as exec/symbol file for gdb
################################################################################

proc file_command {} {
  global Tgdb_cmd

  set file [FileSelectBox -title "Select program to debug" -perm x -sort name]
  if { ![cequal $file ""] } {
    $Tgdb_cmd(file) $file
  }
}

################################################################################
# display the file select box to get a file for use as executable file for gdb
################################################################################

proc exec_command {} {
  global Tgdb_cmd

  set file [FileSelectBox -title "Select an executable file" -perm x -sort name]
  if { ![cequal $file ""] } {
    $Tgdb_cmd(exec-file) $file
  }
}

################################################################################
# display the file select box to get a file for use as symbol file for gdb
################################################################################

proc symbol_command {} {
  global Tgdb_cmd

  set file [FileSelectBox -title "Select a symbol file" -perm x -sort name]
  if { ![cequal $file ""] } {
    $Tgdb_cmd(symbol-file) $file
  }
}

################################################################################
# display the file select box to get a file for use as core file for gdb
################################################################################

proc core_command {} {
  global Tgdb_cmd

  set file [FileSelectBox -title "Select a core file" -perm r -sort name]
  if { ![cequal $file ""] } {
    $Tgdb_cmd(core-file) $file
  }
}

################################################################################
# display the file select box to get a text file to be displayed
################################################################################

proc load_text_command {} {
  global Contents ModTime

  set file [FileSelectBox -title "Select a text file" -perm r -sort name]
  if { ![cequal $file ""] } {
    if { [info exists Contents($file)] } {
      set ModTime($file) 0
    }
    load_file $file 0
  }
}

################################################################################
# save contents of gdb window to a file
################################################################################

proc save_gdb_window {} {
  set file [FileSelectBox -title "Select file for saving gdb window"]
  if { ![cequal $file ""] } {
    if { [catch { set desc [open $file "w"] }] } {
      show_status "Couldn't open file \"$file\"." steady
    } else {
      puts $desc [.f5.text get 1.0 end]
      close $desc
    }
  }
}

################################################################################
# add a directory to the search path for source files
################################################################################

proc add_source_dir {} {
  global Tgdb_cmd

  set file [FileSelectBox -title "Add directory to source path" -file .]
  if { ![cequal $file ""] } {
    if { ![file isdirectory $file] } {
      show_status "No such directory: \"$file\"" steady
      bell
    } else {
      $Tgdb_cmd(directory) $file
    }
  }
}

################################################################################
# change to directory
################################################################################

proc chg_directory {} {
  global Tgdb_cmd

  set file [FileSelectBox -title "Add directory to source path" -file .]
  if { ![cequal $file ""] } {
    if { ![file isdirectory $file] } {
      show_status "No such directory: \"$file\"" steady
      bell
    } else {
      $Tgdb_cmd(cd) $file
    }
  }
}

################################################################################
# see if there is a breakpoint at the line denoted by x,y in window w; if so,
# tell gdb to delete it, otherwise tell him to set one at the first line that
# is known to contain any code
################################################################################

proc toggle_brk { w x y } {
  global ThisFile Tgdb_cmd Cur_thread BptNos BptThread Ttags

  set line [int [$w index @$x,$y]]
  set file [file tail $ThisFile]
  if {   ([lsearch -exact [$w tag names $line.0] brktag] != -1)
      && ([lsearch -exact [$w tag names $line.0] tbrktag] == -1) } {
    $Tgdb_cmd(clear) "$file:$line"
  } elseif { [lsearch -exact [$w tag names $line.0] tbrktag] != -1 } {
    set bpnos ""
    foreach bpt [lsort [array names BptNos]] {
      if {   ($BptThread($bpt) == $Cur_thread)
	  && ($BptNos($bpt) == "$ThisFile,$line")
	  && [info exists Ttags($ThisFile,$Cur_thread)]
          && ([lsearch -exact $Ttags($ThisFile,$Cur_thread) $line.0] != -1) } {
	lappend bpnos $bpt
      }
    }
    if { ![lempty $bpnos] } {
      $Tgdb_cmd(delete) $bpnos
    }
  } else {
    set result [do_dialog "info line $file:$line" silent no no]
    if { ![regexp {^Line .* (starts|is) at (pc|address) (0x[0-9a-f]+)} \
            $result {} {} {} addr] } {
      bell; return
    }
    if { [regexp "but contains no code" $result] } {
      set result [do_dialog "info line *$addr" silent no no]
      if { ![regexp {^Line ([0-9]*) of} $result {} line] } {
        bell; return
      }
    }
    # check again for (t)brktags, since $line might have been redefined
    if { [lsearch -regexp [$w tag names [$w index $line.0]] ^t?brktag] == -1 } {
      if { $Cur_thread == 0 } {
        $Tgdb_cmd(break) "$file:$line"
      } else {
        $Tgdb_cmd(break) "$file:$line thread $Cur_thread"
      }
    } else {
      if { [lsearch -exact [$w tag names [$w index $line.0]] tbrktag] == -1 } {
        $Tgdb_cmd(clear) "$file:$line"
      } else {
        set beg $line.0
	set bpnos ""
	foreach bpt [lsort [array names BptNos]] {
	  if {   ($BptThread($bpt) == $Cur_thread)
	      && ($BptNos($bpt) == "$ThisFile,$line")
	      && [info exists Ttags($ThisFile,$Cur_thread)]
	      && ([lsearch -exact $Ttags($ThisFile,$Cur_thread) $beg] != -1) } {
	    lappend bpnos $bpt
	  }
	}
	if { ![lempty $bpnos] } {
	  $Tgdb_cmd(delete) $bpnos
	}
      }
    }
  }
}

################################################################################
# disassemble the line containing the selection (if owned by the source window)
################################################################################

proc list_disassembly {} {
  global ThisFile Tgdb_cmd

  if {   ([set sel_owner [selection own]] != ".f3.text")
      || ([set sel_range [.f3.text tag ranges sel]] == "") } {
    show_status "No source line selected." steady
    return
  }
  set file [file tail $ThisFile]
  set line [int [lindex $sel_range 0]]
  set result [do_dialog "info line $file:$line" silent no no]
  if { ![regexp {^Line .* (starts|is) at (pc|address) (0x[0-9a-f]+)} \
          $result {} {} {} start_addr] } {
    bell
    return
  }
  if { [regexp "but contains no code" $result] } {
    show_status "Line $line of file \"$file\" contains no code." steady
    return
  }
  if { ![regexp {and ends at (0x[0-9a-f]+)} $result {} end_addr] } {
    bell
    return
  }
  $Tgdb_cmd(disassemble) "$start_addr $end_addr" 
}

################################################################################
#
# PXMon commands
#
################################################################################

add_gdbmon_stuff .f0.f0.file.m

################################################################################
# announce a new task/thread
################################################################################

proc gdbtk_new_task { task } {
  global M_thread Nthreads

  #.f5.text insert end "New task \"$task\""
  #.f5.text yview -pickplace end
  .f0.f0.thrds.m add radiobutton -label $task \
    -variable M_thread -value [incr Nthreads] \
    -command "t_thread $Nthreads"
  .f0.f0.thrds.m entryconfigure $Nthreads -background green
  update idletasks
}

################################################################################
# make $task the current thread
################################################################################

proc gdbtk_switch_to_task { task } {
  global Cur_thread M_thread Tgdb_interactive Proceeding

  set t $task
  for { set i 0 } { $i <= [.f0.f0.thrds.m index last] } { incr i } {
    if { [cequal $t [lindex [.f0.f0.thrds.m entryconfig $i -label] 4]]} {
      set switchto $i
      break
    }
  }
  if { $i > [.f0.f0.thrds.m index last] } {
    set switchto -1
    show_status "Unknown thread <$t>!" 4000
    bell
  } else {
    show_status "\[Switching to \"$task\".\]"
    set Cur_thread $switchto
    set M_thread $switchto
    if { $Proceeding } return
    set inter $Tgdb_interactive
    set Tgdb_interactive 0
    set proceed $Proceeding
    set Proceeding 0
    t_info_line "*\$pc" silent yes
    update_windows new
    set Tgdb_interactive $inter
    set Proceeding $proceed
  }
}

################################################################################
# tell the user that there are pending target messages
################################################################################

proc gdbtk_notify_ptm {} {
  show_status "Received target messages (use \"ptm\" to process them)." steady
  bell
}

################################################################################
# update thread/task list
################################################################################

proc update_tasklist {} {
  global Cur_thread M_thread Nthreads prompt

  set Cur_thread 0
  set M_thread 0
  set Nthreads 0
  if { [.f0.f0.thrds.m index last] > 0 } {
    .f0.f0.thrds.m delete 1 last
  }
  do_dialog "update_tasklist" silent
  append_to_gdbwin "update_tasklist\n$prompt"
}


################################################################################
#
# procs for gdb's command window
#
################################################################################
 
################################################################################
# Tell the user that something went wrong; depending on Tgdb_option(bell) we do
# either make the gdb window flash, or ring the audible bell
################################################################################

proc bell {} {
  global Tgdb_option

  if { $Tgdb_option(bell) == "visual" } {
    incr Tgdb_option(bell,active)
    if { $Tgdb_option(bell,active) > 1 } {
      incr Tgdb_option(bell,active) -1
      return
    }
    set txtcolor [lindex [.f5.text configure -background] 4]
    .f5.text configure -background red
    after 100 ".f5.text configure -background $txtcolor; \
      set Tgdb_option(bell,active) 0"
  } else {
#   set desc [open "/dev/console" "w"]
#   puts -nonewline $desc ""
#   close $desc
#   blt_bell ;# seems to be sexier - now that we use BLT
    tk_bell ;# seems even sexier - now that Tk has it built-in :-)
  }
}

################################################################################
#
# create/delete busy windows while talking to gdb
#
################################################################################

if { "[info commands blt_busy]" == "blt_busy" } {

  proc busy_create { {force 0} } {
    global old_focus Proceeding ForceBusy

    if { $force } {
      set ForceBusy 1
    } elseif { !$Proceeding } {
      return
    }
    if { ![cequal [blt_busy isbusy] ""] } {
      return
    }
    set old_focus(busy) [focus]
    focus .none
    foreach sibling [winfo children .] {
      if { $sibling == ".f4" } {
	foreach but [winfo children .f4] {
	  blt_busy hold $but
	}
	blt_busy forget .f4.stop
      } else {
	blt_busy hold $sibling
      }
    }
    catch {blt_busy forget .dbg}
    update
    bind .f4.stop <Control-c> {.f4.stop invoke; break}
    focus .f4.stop
  }

  proc busy_delete { {force 0} } {
    global old_focus ForceBusy

    if { [info exists ForceBusy] } {
      if { !$force } {
	return
      }
      unset ForceBusy
    }
    if { [cequal [blt_busy isbusy] ""] } {
      return
    }
    focus .none
    bind .f4.stop <Control-c> { break }
    catch {
      foreach bwin [blt_busy isbusy] {
	blt_busy forget $bwin
      }
    }
    catch { focus $old_focus(busy) }
  }

} else {

  proc busy_create { {force 0} } {
    global TkBusy top_left_arrow Proceeding use_dbgwin ForceBusy

    if { $force } {
      set ForceBusy 1
    } elseif { !$Proceeding } {
      set TkBusy ""
      return
    }
    if { [info exists TkBusy] } return
    bind TkBusy <FocusIn> { focus [winfo toplevel %W] }
    set TkBusy "destroy .tkbusy;focus [focus];"
    set script "button .tkbusy;bindtags .tkbusy TkBusy;"
    append script "bindtags .f4.stop \"[bindtags .f4.stop]\";.f4.stop co \
      -cu $top_left_arrow;bind TkBusy <Control-c> \".f4.stop in\";"
    #set w .; set wc [winfo children $w]
    #while { "$wc" != "" } {
    #  append w " $wc"; regsub -all {([^ ]+)} $wc {[winfo ch \1] } wc
    #    set wc [string trim [subst $wc]]}
    set w [info comm .*]
    if { $use_dbgwin } {
      set w [lindex [intersect3 [info comm .dbg.*] $w] 2]
    }
    foreach c $w {append TkBusy "$c con -cu \"[$c cg -cu]\";bindtags $c \
	\"[bindtags $c]\";";bindtags $c TkBusy;$c con -cu watch
	if { "[winfo cla $c]" == "Text" } {foreach t [$c tag nam] {
	foreach b [$c tag bin $t] {
		append TkBusy "$c tag bin $t $b \{[$c tag bin $t $b]\};"
		$c tag bin $t $b { }}}}}
    append TkBusy \
      ".f4.cont co -s normal;" \
      ".f4.next co -s normal;" \
      ".f4.step co -s normal;" \
      ".f4.finish co -s normal;" \
      "catch {.asm.f2.cont co -s normal};" \
      "catch {.asm.f2.next co -s normal};" \
      "catch {.asm.f2.step co -s normal};" \
      "catch {.asm.f2.finish co -s normal};"
    eval $script
  }

  proc busy_delete { {force 0} } {
    global TkBusy ForceBusy

    if { [info exists ForceBusy] } {
      if { !$force } {
	return
      }
      unset ForceBusy
    }
    if { ![info exists TkBusy] } return
    eval $TkBusy; unset TkBusy
  }

}

################################################################################
#
# send a command to gdb and return its output; this also handles queries in
# case gdb wants the user to confirm something, and it automatically handles
# gdb's line information output ("-fullname"), as well as "exceptional output"
#
################################################################################

proc do_dialog {say_what how {showpc no} {loadfile yes} } {
  global Tgdb_busy prompt Tgdb_option std_out QueryResult Proceeding
  global AutoDisps PS2 Nthreads ShowSource Cur_thread M_thread use_expect
  global use_remote use_dbgwin ProgCancelled AsmOpts AsmCachePC
  global switched_to_dbgwin old_focus

  proc filter_buffer { buffer } {
    global Tgdb_interactive
    upvar answer result
    upvar in_disp disp
    upvar skip_line skip_empty_line

    # examine each line of gdb's output to see if it should be directly output
    # to the gdb window, or if it should be saved for later examinations
    set lines [split $buffer "\n"]
    set maxlines [expr [llength $lines] - 1]
    for { set lineno 0 } { 1 } { incr lineno } {
      if { [cequal [set line [lindex $lines $lineno]] ""] } {
	if { $lineno >= $maxlines } {
          return
	}
	if { $skip_empty_line } {
	  set skip_empty_line 0
	  append result "\n"
	}
	if { $Tgdb_interactive } {
	  .f5.text insert end "\n"
	  .f5.text yview -pickplace end
	}
	continue
      }

      # we disabled all auto-displays at the beginning of this dialog, but if
      # a "enable display" command which is attached to a breakpoint re-enabled
      # at least one of them, we shouldd output the lines to the gdb window...
      if { $disp } {
	.f5.text insert end "$line\n"
	.f5.text yview -pickplace end
	update idletasks
	if { [cequal $line "\}"] } {
	  set disp 0
	}
	continue
      }
      if { [regexp {^[0-9]+: } $line] } {
	.f5.text insert end "$line\n"
	.f5.text yview -pickplace end
	update idletasks
	if { [cequal [cindex $line end] "\{"] } {
	  set disp 1
	}
	continue
      }

      if { [regexp "^Reading in symbols for " $line] } {
	continue
      }
      if {   [regexp "^Starting program" $line]
	  || [regexp "^Continuing" $line] } {
	if { $Tgdb_interactive } {
	  .f5.text insert end "$line\n"
	  .f5.text yview -pickplace end
	}
	show_status $line steady
	update idletasks
	append result "$line\n"
	set skip_empty_line 1
	continue
      }
      if {   [regexp "^Program " $line]
	  || [regexp "^Not confirmed" $line] } {
	if { $Tgdb_interactive } {
	  .f5.text insert end "$line\n"
	  .f5.text yview -pickplace end
	} else {
	  append result "\n$line\n"
	}
	continue
      }
      if { $Tgdb_interactive } {
	if { ![regexp [cconcat "\x1a\x1a" {(.:)?[^:]*[:;][0-9]+[:;]([0-9]+)[:;]}] $line]} {
	  .f5.text insert end "$line\n"
	  .f5.text yview -pickplace end
	  update idletasks
	}
	append result "$line\n"
	continue
      }
      if {   [regexp "^The program is not" $line]
          || [regexp "^No selected frame" $line]
          || [regexp "^Value returned " $line]
	  || [regexp [cconcat "\x1a\x1a" {(.:)?[^:]*[:;][0-9]+[:;]([0-9]+)[:;]}] $line]
	  || [regexp {^"finish" not meaningful} $line] } {
        append result "$line\n"
	continue
      }
      if {   [regexp {^0x[0-9a-fA-F]+ in } $line]
          || [regexp {^[a-zA-Z0-9_]+[^)]*\) at .*:[0-9]*.$} $line]
	  || [regexp {^[a-zA-Z0-9_]+[^)]*\) \(.* line [0-9]+\)$} $line]
	  || [regexp {^Breakpoint [0-9]+} $line]
	  || [regexp {^Bpt [0-9]+} $line] } {
	if {   [regexp -indices {Breakpoint [0-9]+} $result {} pos]
	    || [regexp -indices {Bpt [0-9]+} $result {} pos] } {
	  set result [string range $result 0 [lindex $pos 0]]
	  .f5.text insert end "$line\n"
	  .f5.text yview -pickplace end
	}
        append result "$line\n"
	continue
      }

      # no pattern was matched, so just output the line
      .f5.text insert end "$line\n"
      .f5.text yview -pickplace end
      update idletasks
    }
  }

  proc temp_disable_disps {} {
    global prompt AutoDisps use_expect

    # temporarily disable auto-display expressions while proceeding, so that
    # a "continue" command which is attached to a breakpoint won't lead to
    # "endless" display expression output...
    if { $AutoDisps != "" } {
      if { $use_expect } {
	exp_send "disable display\r"
	expect $prompt
      } else {
        catch {gdb_cmd "disable display"}
      }
    }
  }

  proc reenable_disps {} {
    global prompt AutoDisps use_expect
    upvar answer result

    # re-enable auto-display expressions and sample their output
    if { $AutoDisps != "" } {
      if { $use_expect } {
	exp_send "enable display $AutoDisps\r"
	expect $prompt
	exp_send "display\r"
	set first 1
	expect {
	  -re ".*\n" {
	    regsub -all "\r|" $expect_out(buffer) "" expect_out(buffer)
	    if { $first } {
	      set first 0
	      set expect_out(buffer) [crange $expect_out(buffer) \
	        [string first "\n" $expect_out(buffer)]+1 end]
	    }
	    append result $expect_out(buffer)
	    exp_continue;
          }

	  $prompt {
	    if { ![cequal $expect_out(buffer) $prompt] } {
	      append result [crange $expect_out(buffer) 0 end-[clength $prompt]]
	    }
	  }

	  eof {
	    exit_tgdb
	  }
        }
      } else {
        #set result ""
        catch {gdb_cmd "enable display $AutoDisps"}
        catch {gdb_cmd "display"} myresult
        regsub -all "\r|^G" $myresult "" myresult
	append result $myresult
      }
    }
  }

  #################
  # here we go... #
  #################

# if { $use_expect } {
#   expect_background {}
# }
  if { $Tgdb_option(TraceGDB) } { puts $std_out ">>> START OF DO_DIALOG" }
  set Tgdb_busy 1
  busy_create

  set switched_to_dbgwin ""
  if { $Proceeding } {
    temp_disable_disps
    if { $use_dbgwin && $Tgdb_option(RaiseDbgWindow) } {
      proc schedule_dbg_raise { x y move } {
        global switched_to_dbgwin

	after 1000 " \
	  if { \[winfo exists .yesno\] } { \
	    set switched_to_dbgwin \
	      \[after 500 \"schedule_dbg_raise $x $y $move\"\] \
	  } elseif { \$Proceeding } { \
	    raise_debug_window; \
	    if $move { \
	      xaccess eventmotion -widget .dbg.f0.xterm -x $x -y $y \
	    } \
	  } \
	"
      }
      if { $Tgdb_option(WarpToDbgWindow) } {
	set current_x [winfo pointerx .]
	set current_y [winfo pointery .]
	set dbgw [winfo width .dbg.f0.xterm]
	set dbgh [winfo heigh .dbg.f0.xterm]
	set dbgx [expr ([winfo x .dbg.f0.xterm] + $dbgw) / 2]
	set dbgy [expr ([winfo y .dbg.f0.xterm] + $dbgh) / 2]
	set switched_to_dbgwin [schedule_dbg_raise $dbgx $dbgy 1]
      } else {
	set switched_to_dbgwin [schedule_dbg_raise 0 0 0]
      }
    }
  }

  set answer ""
  set first 1
  set QueryResult ""
  set filter_query 0
  set in_disp 0
  set skip_line 0
  set switchto -1
  set ProgCancelled 0
  if { ![cequal $say_what ""] } {
    regsub -all "\t" $say_what " " say_what
    if { $Tgdb_option(TraceGDB) } { puts $std_out "SEND <$say_what>" }
    if { $use_expect } {
      exp_send "$say_what\r"
    } else {
      catch { gdb_cmd $say_what } result
    }
  } else {
    if { $use_remote } {
      catch { gdb_cmd "" } result
    } elseif { !$use_expect } {
      return ""
    }
  }

  if { !$use_expect } {
    regsub -all "\r|" $result "" result
    if { $Tgdb_option(TraceGDB) } {
      if { $use_remote } {
	puts $std_out "RECV <$result$prompt>"
      } else {
	puts $std_out "RECV <$result>"
      }
    }
    if { !$Proceeding } {
      set answer $result
    } else {
      filter_buffer $result
      reenable_disps
    }
  } else {
    expect {
      -re {.*\?.*\(y or n\) } {
	regsub -all "\r|" $expect_out(buffer) "" expect_out(buffer)
	if { $Tgdb_option(TraceGDB) } {
	  puts $std_out "RECV <$expect_out(buffer)>"
	}
	if { $first } {
	  set first 0
	  set expect_out(buffer) [crange $expect_out(buffer) \
	    [string first "\n" $expect_out(buffer)]+1 end]
	}
	if { [yes_no_box "$expect_out(buffer)"] } {
	  append QueryResult y
	  exp_send "y\r"
	} else {
	  append QueryResult n
	  exp_send "n\r"
	}
	expect {
	  -re "\[ny\]\r\n" {
	    # ignore
	  }
	}
	#set filter_query 1
	#set first 1
	#set answer ""
	exp_continue
      }

      "\r\n\x1a\x1aInterrupt\r\n" {
	regsub -all "\r|" $expect_out(buffer) "" expect_out(buffer)
	if { $Tgdb_option(TraceGDB) } {
	  puts $std_out "RECV <$expect_out(buffer)>"
	}
	set expect_out(buffer) [string trimleft $expect_out(buffer)]
	if { [string first \x1a\x1aInterrupt $expect_out(buffer)] != 0 } {
	  set expect_out(buffer) [crange $expect_out(buffer) \
	    0 [string first \x1a\x1aInt $expect_out(buffer)]-1]
	  if { $first } {
	    set first 0
	    set expect_out(buffer) [crange $expect_out(buffer) \
	      [string first "\n" $expect_out(buffer)]+1 end]
	  }
	  if { $filter_query } {
	    if { [regexp "^(y|n)\n" $expect_out(buffer)] } {
	      set filter_query 0
	      set expect_out(buffer) [crange $expect_out(buffer) 2 end]
	    } elseif { [regexp "\n(y|n)\n" $expect_out(buffer)] } {
	      set filter_query 0
	      regsub "\n(y|n)\n" $expect_out(buffer) "\n" expect_out(buffer)
	    }
	  }
	  if { !$Proceeding } {
	    append answer $expect_out(buffer)
	  } else {
	    filter_buffer $expect_out(buffer)
	  }
	}
	bell
	show_status "Received target messages (use \"ptm\" to process them)." \
	  steady
	exp_continue
      }

      -re "^\\\[New (\[^\]\]*)\\\]\r" {
	regsub -all "\r|" $expect_out(buffer) "" expect_out(buffer)
	if { $Tgdb_option(TraceGDB) } {
	  puts $std_out "RECV <$expect_out(buffer)>"
	}
	.f5.text insert end $expect_out(buffer)
	.f5.text yview -pickplace end
	.f0.f0.thrds.m add radiobutton -label $expect_out(1,string) \
	  -variable M_thread -value [incr Nthreads] \
	  -command "t_thread $Nthreads"
	.f0.f0.thrds.m entryconfigure $Nthreads -background green
	exp_continue
      }

      -re "^\\\[Switching to (\[^\]\]*)\\\]\r" {
	regsub -all "\r|" $expect_out(buffer) "" expect_out(buffer)
	if { $Tgdb_option(TraceGDB) } {
	  puts $std_out "RECV <$expect_out(buffer)>"
	}
	set t $expect_out(1,string)
	for { set i 0 } { $i <= [.f0.f0.thrds.m index last] } { incr i } {
	  if { [cequal $t [lindex [.f0.f0.thrds.m entryconfig $i -label] 4]]} {
	    set switchto $i
	    break
	  }
	}
	if { $i > [.f0.f0.thrds.m index last] } {
	  set switchto -1
	  show_status "Unknown thread <$t>!" 4000
	  bell
	}
	if { !$Proceeding } {
	  append answer $expect_out(buffer)
	} else {
	  filter_buffer $expect_out(buffer)
	}
	exp_continue
      }

      -re "(\\\[(\[0-9\])+\\\]\[^\n\]*\n)+$PS2 " {
	regsub -all "\r|" $expect_out(buffer) "" expect_out(buffer)
	if { $Tgdb_option(TraceGDB) } {
	  puts $std_out "RECV <$expect_out(buffer)>"
	}
	if { $first } {
	  set first 0
	  set expect_out(buffer) [crange $expect_out(buffer) \
	    [string first "\n" $expect_out(buffer)]+1 end]
	}
	exp_send [choose_bpt $expect_out(buffer)]\r
	set first 1
	exp_continue
      }

      -re "Loading \[0-9\]+ bytes \\\(text: .*\r\n" {
	regsub -all "\r|" $expect_out(buffer) "" expect_out(buffer)
	if { $first } {
	  set first 0
	  set expect_out(buffer) [crange $expect_out(buffer) \
	    [string first "\n" $expect_out(buffer)]+1 end]
	}
	if { $filter_query } {
	  if { [regexp "^(y|n)\n" $expect_out(buffer)] } {
	    set filter_query 0
	    set expect_out(buffer) [crange $expect_out(buffer) 2 end]
	  } elseif { [regexp "\n(y|n)\n" $expect_out(buffer)] } {
	    set filter_query 0
	    regsub "\n(y|n)\n" $expect_out(buffer) "\n" expect_out(buffer)
	  }
	}
	if { $Tgdb_option(TraceGDB) } {
	  puts $std_out "RECV <$expect_out(buffer)"
	}
	append_to_gdbwin $answer$expect_out(buffer)
	set answer ""
	set first 1
	expect {
	  -re "Program <.*> loaded into target\.\r\n" {
	    regsub -all "\r|" $expect_out(buffer) "" expect_out(buffer)
	    if { $Tgdb_option(TraceGDB) } {
	      puts $std_out "$expect_out(buffer)>"
	    }
	    append_to_gdbwin $expect_out(buffer)
	  }

	  -re {... additional jumps ...} {
	    if { $Tgdb_option(TraceGDB) } {
	      puts -nonewline $std_out "$expect_out(buffer)"
	      flush $std_out
	    }
	    append_to_gdbwin $expect_out(buffer)
	    exp_continue
	  }

	  -re "Quit\r\n" {
	    if { $Tgdb_option(TraceGDB) } {
	      puts "Quit>"
	    }
	    append_to_gdbwin "Quit\n"
	  }

	  -re {[tdb]} {
	    if { $Tgdb_option(TraceGDB) } {
	      puts -nonewline $std_out "$expect_out(buffer)"
	      flush $std_out
	    }
	    append_to_gdbwin $expect_out(buffer)
	    exp_continue
	  }

	  eof {
	    if { $Tgdb_option(TraceGDB) } {
	      puts $std_out "RECV \[EOF\]>"
	    }
	    exit_tgdb
	  }
	}
	exp_continue
      }

      -re ".*\r\n\r.*\n" {
	regsub -all "\r\n\r" $expect_out(buffer) "" expect_out(buffer)
	regsub -all "\r|" $expect_out(buffer) "" expect_out(buffer)
	if { $Tgdb_option(TraceGDB) } {
	  puts $std_out "RECV <$expect_out(buffer)>"
	}
	if { $first } {
	  set first 0
	  set expect_out(buffer) [crange $expect_out(buffer) \
	    [string first "\n" $expect_out(buffer)]+1 end]
	}
	if { $filter_query } {
	  if { [regexp "^(y|n)\n" $expect_out(buffer)] } {
	    set filter_query 0
	    set expect_out(buffer) [crange $expect_out(buffer) 2 end]
	  } elseif { [regexp "\n(y|n)\n" $expect_out(buffer)] } {
	    set filter_query 0
	    regsub "\n(y|n)\n" $expect_out(buffer) "\n" expect_out(buffer)
	  }
	}
	if { !$Proceeding } {
	  append answer $expect_out(buffer)
	} else {
	  filter_buffer $expect_out(buffer)
	}
	exp_continue;
      }

      -re ".*\n" {
	regsub -all "\r|" $expect_out(buffer) "" expect_out(buffer)
	if { $Tgdb_option(TraceGDB) } {
	  puts $std_out "RECV <$expect_out(buffer)>"
	}
	if { $first } {
	  set first 0
	  set expect_out(buffer) [crange $expect_out(buffer) \
	    [string first "\n" $expect_out(buffer)]+1 end]
	}
	if { $filter_query } {
	  if { [regexp "^(y|n)\n" $expect_out(buffer)] } {
	    set filter_query 0
	    set expect_out(buffer) [crange $expect_out(buffer) 2 end]
	  } elseif { [regexp "\n(y|n)\n" $expect_out(buffer)] } {
	    set filter_query 0
	    regsub "\n(y|n)\n" $expect_out(buffer) "\n" expect_out(buffer)
	  }
	}
	if { !$Proceeding } {
	  append answer $expect_out(buffer)
	} else {
	  filter_buffer $expect_out(buffer)
	}
	exp_continue;
      }

      $prompt {
	if { $Tgdb_option(TraceGDB) } {
	  puts $std_out "RECV <$expect_out(buffer)>"
	}
	if { ![cequal $expect_out(buffer) $prompt] } {
	  append answer [crange $expect_out(buffer) 0 end-[clength $prompt]]
	}
	if { $Proceeding } {
	  reenable_disps
	}
      }

      eof {
	if { $Tgdb_option(TraceGDB) } {
	  puts $std_out "RECV <$expect_out(buffer)><\[EOF\]>"
	}
	exit_tgdb
      }
    }
  }

  if { $Proceeding } {
    catch { unset AsmCachePC }
    if { "$switched_to_dbgwin" != "" } {
      if { [catch {after info $switched_to_dbgwin}] } {
	if { ![string match "n*" "$QueryResult"] } {
	  if { [cequal $old_focus(busy) .asm] } {
	    raise_asm_window
	  } else {
	    raise_main_window
	  }
          if { $Tgdb_option(WarpToDbgWindow) && !$ProgCancelled } {
	    xaccess eventmotion -x $current_x -y $current_y"]
	  }
	}
      } else {
	after cancel $switched_to_dbgwin
      }
    }
  }

  if { $loadfile == "yes" } {
    # don't display any sources if we stopped at a
    # place where no debug info is available...
    if {   ($showpc != "no") && ($switchto < 0) && ($Proceeding || $ShowSource)
	&& ![regexp \
	      [cconcat "\x1a\x1a" {(.:)?[^:]*[:;][0-9]+[:;][0-9]+[:;](beg|middle):0x[0-9a-f]+}] \
	      $answer]
       } {
      set file ""
      if { $use_expect } {
        exp_send "info line *\$pc\r"
        expect $prompt {
	  if { ( ![regexp \
	     [cconcat "\x1a\x1a" {((.:)?[^:]*)[:;][0-9]+[:;][0-9]+[:;](beg|middle):0x[0-9a-f]+}] \
	            $expect_out(buffer) {} file]
	    && ![regexp {Line [0-9]+ of "(.*)" .* at (pc|address) 0x[0-9a-f]+} \
		    $expect_out(buffer) {} file])
	    || (   ![file readable $file]
	        && ![file readable [file2nfspath $file]]) } {
	    global Line ThisFile LastFile LastLine LastThread Cur_thread

	    if { $ThisFile != "" } {
	      set LastFile $ThisFile
	      set LastLine [expr [int [.f3.text index @0,0]] - 1]
	      set LastThread $Cur_thread
	    }
	    .f3.text configure -state normal
	    .f3.text delete 1.0 end
	    .f3.text configure -state disabled
	    set Line ""
	    set ThisFile ""
          }
        }
      } else {
        catch { gdb_cmd "info line *\$pc" } result
	if { ( ![regexp \
	     [cconcat "\x1a\x1a" {((.:)?[^:]*)[:;][0-9]+[:;][0-9]+[:;](beg|middle):0x[0-9a-f]+}] \
		  $result {} file]
	    && ![regexp {Line [0-9]+ of "(.*)" .* at (pc|address) 0x[0-9a-f]+} \
		  $result {} file])
	    || (   ![file readable $file]
	        && ![file readable [file2nfspath $file]]) } {
	  global Line ThisFile LastFile LastLine LastThread Cur_thread

	  if { $ThisFile != "" } {
	    set LastFile $ThisFile
	    set LastLine [expr [int [.f3.text index @0,0]] - 1]
	    set LastThread $Cur_thread
	  }
	  .f3.text configure -state normal
	  .f3.text delete 1.0 end
	  .f3.text configure -state disabled
	  set Line ""
	  set ThisFile ""
        }
      }
    }
    while { [regexp \
       [cconcat "\x1a\x1a" {((.:)?[^:]*)[:;]([0-9]+)[:;]([0-9]+)[:;](beg|middle):(0x[0-9a-f]+)}] \
	      $answer {} file {} line off where pc] } {
      if { $showpc == "no" } {
        load_file $file $line no
      } else {
        load_file $file $line yes
      }
      if { [info exists AsmOpts(cache)] && $AsmOpts(cache) } {
	set AsmCachePC $pc
      }
      regsub \
	[cconcat "\x1a\x1a" {(.:)?[^:]*[:;][0-9]+[:;][0-9]+[:;](beg|middle):0x[0-9a-fA-F]*.}] \
	$answer "" answer
    }
  }
  regsub -all "Reading in symbols for .*done.\n" $answer "" answer
  if { ($how == "verbose") && !$Proceeding && (![cequal $answer ""]) } {
    .f5.text insert end $answer
    .f5.text yview -pickplace end
  }
  if { $Tgdb_option(TraceGDB) } { puts $std_out "<<< END OF DO_DIALOG" }
  busy_delete

# expect_background {
#   "\r\n\x1a\x1aInterrupt\r\n" {
#     #exp_send ""
#     #expect "!$prompt"
#     bell
#     show_status "Received target messages (use \"ptm\" to process them)." \
#       steady
#   }
# }
  # avoid Enter/Leave event side effects created by busy windows, i.e. give
  # buttons a chance to redraw themselves before new actions can take place
  after 150 {set Tgdb_busy 0}
  # handle thread switches
  if { $switchto != -1 } {
    global Tgdb_interactive

    set Cur_thread $switchto
    set M_thread $switchto
    set inter $Tgdb_interactive
    set Tgdb_interactive 0
    set proceed $Proceeding
    set Proceeding 0
    t_info_line "*\$pc" silent yes
    update_windows new
    set Tgdb_interactive $inter
    set Proceeding $proceed
  }
  return $answer
}

################################################################################
#
# evaluate a gdb command and filter gdb's output; must only be called for
# commands typed on the command line (not for commands produced by buttons,
# unless they simulate user interaction visibly)
#
################################################################################

proc filter_output { cmd {arg ""} } {
  global Tgdb_cmd Tgdb_interactive prompt

  set Tgdb_interactive 1
  if { [info exists Tgdb_cmd($cmd)] } {
    $Tgdb_cmd($cmd) $arg
  } else {
    if { ![cequal $arg ""] } {
      append cmd " $arg"
    }
    do_dialog $cmd verbose yes
  }
  .f5.text insert end $prompt
  .f5.text yview -pickplace insert
  set Tgdb_interactive 0
  update
}

################################################################################
#
# exit from tgdb (write tgdb init file and destroy all windows)
#
################################################################################

proc exit_tgdb { args } {
  global xterm_pid QueryResult ExitTgdb Tgdb_interactive prompt
  global Tgdb_option WinPos WinSize Tgdb_startfile WinProc AsmOpts
  global gdb_history gdb_history_nr gdb_history_max gdb_history_spot
  global use_expect use_remote use_dbgwin debugger CpuOpts

  if { $use_expect || $use_remote } {
    if { ![info exists ExitTgdb] } { set ExitTgdb 0 }
    # avoid nested calling (due to EOF when killing gdb in do_dialog)
    if { $use_expect } {
      if { [incr ExitTgdb] > 1 } { incr ExitTgdb -1; return }
    } elseif { [incr ExitTgdb] == 3 } {
      return
    }

    set QueryResult ""
    if { ![catch {do_dialog "quit" silent}] && ($QueryResult == "n") } {
      if { $use_remote } {
	set result "Not confirmed.\n"
	if { $Tgdb_interactive } {
	  append_to_gdbwin $result
	} else {
	  append_to_gdbwin "quit\n$result$prompt"
	}
      }
      set ExitTgdb 0
      return
    }
  } else { ;# gdbtk
    if { "$debugger" != "gdb166" } {
      set result [do_dialog "info program" silent]
      if { ![regexp {^The program being debugged is not being r.*} $result] } {
        if { ![yes_no_box \
	        "The program is running.  Quit anyway (and detach it)?"] } {
	  if { $Tgdb_interactive } {
	    append_to_gdbwin "Not confirmed.\n"
	  } else {
	    append_to_gdbwin "quit\nNot confirmed.\n$prompt"
	  }
	  return
	}
      }
    } else {
      if { ![info exists ExitTgdb] } { set ExitTgdb 0 }
      if { [incr ExitTgdb] == 1 } {
	set result [do_dialog "quit" silent]
	if { [cequal $result "Not confirmed.\n"] } {
	  if { $Tgdb_interactive } {
	    append_to_gdbwin $result
	  } else {
	    append_to_gdbwin "quit\n$result$prompt"
	  }
	  set ExitTgdb 0
	  return
        }
      }
    }
    # redefine gdbtk_tcl_query so we won't be called again
    proc gdbtk_tcl_query { args } { return 1 }
  }
  show_status "Thanks for using tgdb. Good Bye."
  update idletasks

  if { $use_dbgwin } {
    catch {close -slave}
    catch {kill $xterm_pid}
    catch {.shell.f0.xterm stoptksteal}
  }

  if { $Tgdb_option(SaveOptions) } {
    catch {
      set desc [open $Tgdb_startfile "w"]
      puts $desc "#\n# tgdb startup file\n#\n"
      puts $desc "global WinPos Tgdb_option AsmOpts CpuOpts\n"
      puts $desc "# options"
      foreach opt [array names Tgdb_option] {
	puts $desc "set Tgdb_option($opt) $Tgdb_option($opt)"
      }
      if { [info exists AsmOpts(size)] } {
	puts $desc "set AsmOpts(size) $AsmOpts(size)"
	puts $desc "set AsmOpts(mode) $AsmOpts(mode)"
	puts $desc "set AsmOpts(cache) $AsmOpts(cache)"
	puts $desc "set AsmOpts(status) $AsmOpts(status)"
      }

      # remember selected registers and their order
      puts $desc "\n# cpu register settings"
      if { [info exists CpuOpts(orig_regs)] } {
	puts $desc "set CpuOpts(orig_regs) \[list \\"
	foreach reg $CpuOpts(orig_regs) {
	  puts $desc "  $reg \\"
	}
	puts $desc "\]"
	puts $desc "set CpuOpts(sel_regs) \[list \\"
	foreach reg $CpuOpts(sel_regs) {
	  puts $desc "  $reg \\"
	}
	puts $desc "\]"
      }

      # remember geometry of previously/currently opened windows
      set windows ""
      foreach win [winfo children .] {
	set top [winfo toplevel $win]
	if { [lsearch -exact $windows $top] == -1 } {
	  lappend windows $top
	  set WinPos($top) [wm geometry $top]
	}
      }
      puts $desc "\n# window geometries"
      foreach win [array names WinPos] {
	puts $desc "set WinPos($win) \"$WinPos($win)\""
      }
      # remember gdb window's height
      set txtb [lindex [.f5.text conf -bd] 4]
      set txth $WinSize(.,y2)
      puts $desc "set WinGdb [expr ([winfo height .f5.text] - $txtb) / $txth]"
      # remember which windows are currently opened
      if { $windows != "" } {
        puts $desc "\n# open windows"
	foreach win $windows {
	  if { [lsearch -exact \
		 ". .about .dbg .help .thelp .pxmon .pxmgfx" $win] != -1 } {
	    continue
	  }
	  puts $desc "catch \{$WinProc($win)\}"
	}
      }
      # should we save tgdb's command history?
      if { $Tgdb_option(SaveHistory) } {
	puts $desc "\n# command history"
	puts $desc "global gdb_history gdb_history_max"
	puts $desc "global gdb_history_nr gdb_history_spot\n"
	puts $desc "set gdb_history_max $gdb_history_max"
	puts $desc "set gdb_history_nr $gdb_history_nr"
	puts $desc "set gdb_history_spot $gdb_history_spot"
	loop i 0 $gdb_history_max {
	  if { ![info exists gdb_history($i)] } break
	  puts $desc "set gdb_history($i) \{$gdb_history($i)\}"
	}
      }
      close $desc
    }
  }
  if { $use_expect || $use_remote } {
    catch {destroy .}
    exit 0
  } elseif { "$debugger" != "gdb166" } {
    do_dialog "quit" silent
  }
}

################################################################################
#
# initialize/reset source window
#
################################################################################

proc init_srcwin {} {
  global Mark Line ThisFile Contents TextBindings LentryBindings SearchBindings
  global LastFile LastLine colormodel Cur_thread M_thread Nthreads LastThread

  if { [cequal [bind .f3.text <1>] " "] } {
    set save_bindings 0
  } else {
    set save_bindings 1
  }
  .f3.text configure -state normal
  .f3.text delete 1.0 end
  .f3.text configure -state disabled
  catch {unset Contents}
  set ThisFile ""
  set Line ""
  set LastFile ""
  set LastLine 0
  set LastThread 0
  set Cur_thread 0
  set M_thread 0
  set Nthreads 0
  if { [.f0.f0.thrds.m index last] > 0 } {
    .f0.f0.thrds.m delete 1 last
  }
  if { $save_bindings } {
    foreach bind [bind .f3.text] {
      set TextBindings($bind) [bind .f3.text $bind]
      bind .f3.text $bind { }
    }
    foreach bind [bind .title.f1.lentry] {
      set LentryBindings($bind) [bind .title.f1.lentry $bind]
      bind .title.f1.lentry $bind { }
    }
    foreach bind [bind .f2.search] {
      set SearchBindings($bind) [bind .f2.search $bind]
      bind .f2.search $bind { }
    }
  }
  if { $colormodel == "color" } {
    .f3.text tag configure mytag -background green
  } else {
    .f3.text tag configure mytag -foreground white -background black
  }
  del_all_brktags
}

################################################################################
################################################################################
#
# "main()"
#
################################################################################
################################################################################

# initialize global variables
set FreezeStatus 0
set SourcePath [pwd]
set gdb_history_max 128
set gdb_history(0) ""
set gdb_history_nr 0
set gdb_history_spot 0
set Tgdb_option(RemCRs) 0
set Tgdb_option(ProgPath) 1
set Tgdb_option(SaveHistory) 1
set Tgdb_option(bell) visual
set Tgdb_option(bell,active) 0
set Tgdb_option(RaiseWindows) 0
set Tgdb_option(gdbRO) 0
set Tgdb_option(TraceGDB) 0
set Tgdb_option(UpdateBpts) 0
set Tgdb_option(ClassKludge) 0
set Tgdb_option(SaveOptions) 1
set Tgdb_option(RaiseDbgWindow) 1
set Tgdb_option(WarpToDbgWindow) 1
set Tgdb_busy 0
set Tgdb_interactive 0
set ShowSource 0
set BlinkingPC 0
set BlinkingAsmPC 0
set Proceeding 0
set AutoDisps ""
set timeout -1
set Cur_thread 0
set M_thread 0
set Nthreads 0
set MemScrollInProgress 0
set AsmScrollInProgress 0
set CmdInProgress 0
catch {rename bitmap blt_bitmap}
if { "[info commands blt_bitmap]" == "blt_bitmap" } {
  blt_bitmap define stipple_pat { {4 4} {08 04 02 01} }
} else {
  image create bitmap stipple_pat -data {
    #define stipple_pat_width 4
    #define stipple_pat_height 4
    static char stipple_pat_bits[] = {
      0x08, 0x04, 0x02, 0x01};
  }
}

################################################################################
#
# initialize tgdb/gdb interface
#
################################################################################

if { $use_expect || $use_remote || $use_dbgwin } {
  if { "$debugger" == "gdb166" } {
    if { $use_expect } {
      set gdb_invocation "$debugger -nw -nx -fullname [split $gdb_argv]"
      set gdb_pid [eval spawn -noecho $dbgrsh $gdb_invocation]
    } else {
      if { $use_socket } {
        set gdb_invocation "$debugger -nw -nx [split $gdb_argv]"
        srv_open_socket
	after 120000 { set gdb_desc -1 }
	eval exec $dbgrsh $gdb_invocation &
	vwait gdb_desc
	after cancel { set gdb_desc -1 }
	if { [cequal $gdb_desc -1] } {
	  puts "Couldn't connect to $debugger."
	  exit 1
	}
      } else {
        set gdb_invocation "$debugger -nw -nx -fullname -notty [split $gdb_argv]"
        set gdb_desc [eval open \"| $dbgrsh $gdb_invocation 2>@ $std_out\" RDWR]
        fconfigure $gdb_desc -blocking 0 -buffering none
        fileevent $gdb_desc readable handle_remote_input
        set gdb_pid [pid $gdb_desc]
      }
    }
    do_dialog "" silent
    if { $gdb16x } {
      do_dialog "set height 0" silent
      do_dialog "set width 0" silent
      do_dialog "set print elements 200" silent
      do_dialog "set print pretty" silent
    } else {
      do_dialog "set screen 0 0" silent
      do_dialog "set array-max 200" silent
      do_dialog "set pretty" silent
    }
  } else {
    # restore original environment variables, so we can also debug programs
    # which use Tcl/Tk and friends by themselves
    catch {set myenv(LD_LIBRARY_PATH) $env(LD_LIBRARY_PATH)}
    catch {set myenv(LIBPATH) $env(LIBPATH)}
    catch {set myenv(TCL_LIBRARY) $env(TCL_LIBRARY)}
    catch {set myenv(TK_LIBRARY) $env(TK_LIBRARY)}
    catch {set myenv(BLT_LIBRARY) $env(BLT_LIBRARY)}
    catch {set myenv(TCLX_LIBRARY) $env(TCLX_LIBRARY)}
    catch {set myenv(TKSTEAL_LIBRARY) $env(TKSTEAL_LIBRARY)}
    catch {set env(LD_LIBRARY_PATH) $env(OLD_LD_LIBRARY_PATH)}
    catch {set env(LIBPATH) $env(OLD_LIBPATH)}
    catch {set env(TCL_LIBRARY) $env(OLD_TCL_LIBRARY)}
    catch {set env(TK_LIBRARY) $env(OLD_TK_LIBRARY)}
    catch {set env(BLT_LIBRARY) $env(OLD_BLT_LIBRARY)}
    catch {set env(TCLX_LIBRARY) $env(OLD_TCLX_LIBRARY)}
    catch {set env(TKSTEAL_LIBRARY) $env(OLD_TKSTEAL_LIBRARY)}

    if { $use_dbgwin } {
      toplevel .dbg -cursor $top_left_arrow
      wm withdraw .dbg
      frame .dbg.f0 -border 4 -relief sunken
      if { [catch {set debfont [option get .dbg.f0 font Font]}] } {
        set debfont ""
      } elseif { $debfont != "" } {
        set debfont "-fn $debfont"
      }
      if { [catch {set debcolor [option get .dbg.f0 foreground Foreground]}] } {
        set debcolor ""
      } elseif { $debcolor != "" } {
        set debcolor "-xrm \"xterm*foreground:$debcolor\""
      }
      set xterm_desc [open "|mkxtty -iconic #+10000+10000 \
	-title \"Tgdb debugee\" \
	-xrm \"xterm*allowSendEvents:true\" $debcolor \
	-xrm \"xterm*background:#d9d9d9\" $debfont -geometry 200x100+0+0"]
      set xterm_line [gets $xterm_desc]
      set xterm_xid [lindex $xterm_line 0]
      set xterm_tty [lindex $xterm_line 1]
      if { ![scan [lindex $xterm_line 2] "%d" xterm_pid] } {
	puts $std_out "Couldn't read xterm_pid (maybe a problem with mkxtty?)."
	catch {.shell.f0.xterm stoptksteal}
	exit
      }
      catch {close $xterm_desc}
      tksteal .dbg.f0.xterm -name "Tgdb debugee" -windowid $xterm_xid
      pack .dbg.f0 -expand 1 -fill both
      pack .dbg.f0.xterm -in .dbg.f0 -expand 1 -fill both
      wm title .dbg "Tgdb debugee window"
      bindtags .dbg .dbg
    }
    #
    # now call gdb
    #
    if { $use_expect || $use_remote } {
      if { $use_dbgwin } {
	catch {set orig_shell $env(SHELL)}
	set env(SHELL) callcld
	set env(TGDB_TTYDEV) $xterm_tty
      }
      if { $use_expect } {
	set gdb_invocation "$debugger -nw -nx -fullname [split $gdb_argv]"
	set gdb_pid [eval spawn -noecho $dbgrsh $gdb_invocation]
      } else {
	if { $use_socket } {
	  set gdb_invocation "$debugger -nw -nx [split $gdb_argv]"
          srv_open_socket
	  after 120000 { set gdb_desc -1 }
	  eval exec $dbgrsh $gdb_invocation &
	  vwait gdb_desc
	  after cancel { set gdb_desc -1 }
	  if { [cequal $gdb_desc -1] } {
	    puts "Couldn't connect to $debugger."
	    exit 1
	  }
	} else {
	  set gdb_invocation "$debugger -nw -nx -fullname -notty [split $gdb_argv]"
	  append gdb_invocation " -tty [exec tty]"
	  set gdb_desc [eval open \"| $gdb_invocation 2>@ $std_out\" RDWR]
	  fconfigure $gdb_desc -blocking 0 -buffering none
	  fileevent $gdb_desc readable handle_remote_input
	  set gdb_pid [pid $gdb_desc]
	}
      }
      catch {
	set env(SHELL) $orig_shell
	unset env(TGDB_TTYDEV) xterm_desc xterm_line debcolor debfont orig_shell
      }
    } else {
      gdb_cmd "tty $xterm_tty"
      gdb_cmd "set env TGDB_TTYDEV=$xterm_tty"
      gdb_cmd "set env TGDB_TTYPID=$xterm_pid"
      set env(SHELL) callcld
      unset xterm_desc xterm_line debcolor debfont
    }
    do_dialog "" silent
    do_dialog "set height 0" silent
    do_dialog "set width 0" silent
    do_dialog "set print elements 200" silent
    do_dialog "set print pretty" silent
    # restore environment variables
    catch {set env(LD_LIBRARY_PATH) $myenv(LD_LIBRARY_PATH)}
    catch {set env(LIBPATH) $myenv(LIBPATH)}
    catch {set env(TCL_LIBRARY) $myenv(TCL_LIBRARY)}
    catch {set env(TK_LIBRARY) $myenv(TK_LIBRARY)}
    catch {set env(BLT_LIBRARY) $myenv(BLT_LIBRARY)}
    catch {set env(TCLX_LIBRARY) $myenv(TCLX_LIBRARY)}
    catch {set env(TKSTEAL_LIBRARY) $myenv(TKSTEAL_LIBRARY)}
    catch {unset myenv}
  }
  catch {unset gdb_invocation}
} else { ;# gdbtk
  do_dialog "" silent
  do_dialog "set height 0" silent
  do_dialog "set width 0" silent
  do_dialog "set print elements 100" silent
  do_dialog "set print pretty" silent
}
destroy .welcome

################################################################################
#
# try to load tgdb's startup file "./.t"$debugger"init" or "~/.t"$debugger"init"
# (or as specified in the TGDB_INITFILE environment variable)
#
################################################################################

if { [info exists env(TGDB_INITFILE)] } {
  set Tgdb_startfile $env(TGDB_INITFILE)
} else {
  set startfile0 ""; append startfile0 "./.t" $debugger "init"
  set startfile1 ""; append startfile1 "~/.t" $debugger "init"
  if { [file readable $startfile0] } {
    set Tgdb_startfile $startfile0
  } else {
    set Tgdb_startfile $startfile1
  }
  unset startfile0 startfile1
}
catch {source $Tgdb_startfile}

################################################################################
#
# show debugee's window
#
################################################################################

if { [winfo exists .dbg] } {
  foreach bind [bind all] {
    bind .dbg $bind [bind all $bind]
  }
  catch {unset bind}
  bind .dbg <Control-c> {
    .f4.stop invoke
    if { !$Tgdb_option(RaiseDbgWindow) } {
      raise_main_window
    }
    break
  }
  if { [catch {wm geometry .dbg $WinPos(.dbg)}] } {
    wm geometry .dbg 570x335+0+0
  }
  wm minsize .dbg 10 10
  wm protocol .dbg WM_DELETE_WINDOW {bell}
  wm protocol .dbg WM_TAKE_FOCUS {focus .dbg}
  wm deiconify .dbg
}

################################################################################
#
# show main window
#
################################################################################

wm title . "tgdb"
wm group . .
wm protocol . WM_TAKE_FOCUS {
  if { $Proceeding } {
    focus .f4.stop
  } else {
    focus .f5.text
  }
}
wm protocol . WM_DELETE_WINDOW {exit_tgdb}
bind . <Visibility> {check_visibility %W}
if { [catch {wm geometry . $WinPos(.)}] } {
  wm geometry . 80x15+100+100
}
wm minsize . 80 5
wm deiconify .
focus .none
tkwait visibility .
update
set txtb [lindex [.f3.text conf -bd] 4]
scan [wm geometry .] "%dx%d" txtw txth
set WinSize(.,x) [expr ([winfo width .f3.text] - $txtb) / $txtw]
set WinSize(.,y) [expr ([winfo height .f3.text] - $txtb) / $txth]
set txtw [lindex [.f5.text conf -width] 4]
set txtb [lindex [.f5.text conf -bd] 4]
set txth [lindex [.f5.text conf -height] 4]
set WinSize(.,x2) [expr ([winfo width .f5.text] - $txtb) / $txtw]
set WinSize(.,y2) [expr ([winfo height .f5.text] - $txtb) / $txth]
unset txtw txtb txth
if { [info exists WinGdb] } {
  .f5.text configure -height $WinGdb
}

################################################################################
#
# initialize windows, commands and help facility
#
################################################################################

show_status "Initializing tgdb..." steady
update idletasks
init_srcwin
init_gdb_help
init_tgdb_cmds
#show_copyright_and_license_policy

################################################################################
#
# gdbtk interface (mostly dummy procs)
#
################################################################################

if { !$use_expect && !$use_remote } {
  proc gdbtk_tcl_fputs { args } {
  # global std_out

  # puts $std_out "gdbtk_tcl_fputs <$args>"
  }
  proc gdbtk_tcl_fputs_error { args } {
  # global std_out

  # puts $std_out "gdbtk_tcl_fputs_error <$args>"
  }
  proc gdbtk_tcl_flush {} {
    update idletasks
  }
  proc gdbtk_tcl_breakpoint { action bpnum } {
  # puts "gdbtk_tcl_breakpoint <$action> <$bpnum>"
  }
  proc gdbtk_tcl_busy {} {
  # puts "gdbtk_tcl_busy"
  }
  proc gdbtk_tcl_idle {} {
  # puts "gdbtk_tcl_idle"
  }
  proc gdbtk_tcl_query { message } {
    return [yes_no_box $message]
  }
}

################################################################################
#
# add path names for exec/symbol/core files; call FileSelectBox (if req'd),
# and list main() function if no core file was specified
#
################################################################################

if { ("$debugger" == "gdb166") && !$gdb16x } {
  #filter_output "info version"
  filter_output "show version"
} else {
  filter_output "show version"
}

if { $Tgdb_option(ProgPath) } {
  if { ![cequal $gdb_symfile ""] } {
    set Tgdb_cwd [file dirname $gdb_symfile]
    $Tgdb_cmd(cd) $Tgdb_cwd
  }
}

$Tgdb_cmd(directory) $Tgdb_cwd
foreach file "$gdb_execfile $gdb_symfile $gdb_corefile" {
  set path [file dirname $file]
  if { ![cequal $path ""] && ![cequal $path "."] } {
    $Tgdb_cmd(directory) "$path"
  }
}
catch {unset file}
catch {unset path}

if { !$use_expect && !$use_remote } {
  if { [cequal $gdb_execfile $gdb_symfile] && ![cequal $gdb_execfile ""] } {
    $Tgdb_cmd(file) $gdb_execfile
  } elseif { ![cequal $gdb_execfile ""] } {
    $Tgdb_cmd(exec-file) $gdb_execfile
    $Tgdb_cmd(symbol-file) $gdb_symfile
  }
  if { ![cequal $gdb_corefile ""] } {
    $Tgdb_cmd(core-file) $gdb_corefile
  }
} else {
  if { ![cequal $gdb_execfile ""] && ![file readable $gdb_execfile] } {
    append_to_gdbwin "$gdb_execfile: No such file or directory.\n\n$prompt"
  }
  if { ![cequal $gdb_symfile $gdb_execfile] && ![cequal $gdb_symfile ""] } {
    if { ![file readable $gdb_symfile] } {
      append_to_gdbwin "$gdb_symfile: No such file or directory.\n\n$prompt"
    }
  }
}
if {   [cequal $gdb_execfile ""]
    && [cequal $gdb_corefile ""]
    && [cequal $gdb_attach_pid ""]
    && [cequal $gdb_symfile ""]
    && [cequal $gdb_cmdfile ""]
    && ($gdb_nx || ![file readable $gdb_initfile]) } {
  file_command
}

################################################################################
#
# load core file or attach to a given process; see description of proc
# process_gdb_args (in gdb_help.tcl) for the reason why tgdb has to do
# this itself instead of just passing these args when gdb is invoked
#
################################################################################

if { [cequal $gdb_corefile ""] && [cequal $gdb_attach_pid ""] } {
  do_dialog "set \$pc=main" silent
  $Tgdb_cmd(list) "main,"
} else {
  if { ![cequal $gdb_corefile ""] } {
    $Tgdb_cmd(core-file) $gdb_corefile
  }
  if { ![cequal $gdb_attach_pid ""] } {
    $Tgdb_cmd(attach) $gdb_attach_pid
  }
}

################################################################################
#
# try to load gdb's startup file(s)
#
################################################################################

if { !$gdb_nx && [file readable $gdb_initfile] } {
  show_status "Loading $gdb_initfile..." steady
  update idletasks
  init_srcwin
  $Tgdb_cmd(source) $gdb_initfile
  show_status ""
}
if { "$gdb_cmdfile" != "" } {
  if { [file readable $gdb_cmdfile] } {
    show_status "Loading command file \"$gdb_cmdfile\"..." steady
    update idletasks
    init_srcwin
    $Tgdb_cmd(source) $gdb_cmdfile
    show_status "Good luck!" 4000
    update idletasks
  } else {
    show_status "Couldn't load command file \"$gdb_cmdfile\"..." 4000
    bell
  }
} else {
  show_status "Good luck!" 4000
  update idletasks
}

if { $colormodel == "monochrome" } {
  blink_line
}

if { $use_dbgwin && !$use_expect && !$use_remote } {
  raise_main_window
}

################################################################################
#
# now let's bite the bullet...
#
################################################################################

focus .f5.text

### EOF ########################################################################
