/*	@(#)rb.h	1.1	1/4/93	(DM)

Headerfile zur Handhabung generischer Ringpuffer

Allgemeine Voraussetzungen:

	Der Header definiert Funktionen zur Definition von Ringpuffern und
	zum Zugriff darauf.

	Die Funktionen setzen voraus, dass Einfuegeoperationen nicht
	von Einfuegeoperationen und Austragungsoperationen nicht von
	Austragungsoperationen unterbrochen werden. Demgegebenueber
	duerfen Einfuege- von Austragungsoperatioen und Austrage-
	von Einfuegeoperation unterbrochen werden.

	Der Test auf "Leeren Ringpuffer" darf "leer" zurueckgeben, wenn
	gerade eine Einfuegeoperation durchgefuehrt wird. Austrageoperationen
	und der Test auf "Leeren Ringpuffer" duerfen sich nicht ueberschneiden.

	Der Test aus "vollen Ringpuffer" darf "voll" zurueckgegeben,
	wenn gerade eine Austrageoperation stattfindet. Der Test
	auf einen vollen Ringpuffer und Einfuegeoperationen duerfen
	sich nicht ueberschneiden.


Funktionsumfang:

	RB_CREATE(name,typ)	deklariert eine Ringpufferdatenstruktur mit
				Name "name" fuer Komponenten von Typ "typ".

	RB_INIT(&rb,start,anzahl)
				initialisiert "rb". "start" wird als
				Speicherbereich fuer "anzahl" Komponenten
				des Ringpuffers angesehen.

	empty= RB_EMPTY(&rb)	Test auf "leeren Ringpuffer"

	full= RB_FULL(&rb)	Test auf "vollen Ringpuffer"

	err= RB_PUSH(&rb,v)	Einfuegen von "v" nach "rb".
				err==1: Puffer voll; err==0: Erfolg

	err= RB_POP(&rb,&v)	Entfernt Element aus "rb" und schreibt es
				nach "v".
				err==1: Puffer leer; err==0: Erfolg

Ergaenzungen FB 03.09.91
	
	wpos= RB_WPOS(&rb)	Liefert die aktuelle Schreibposition
				des Ringpuffers.

	RB_WRESET(&rb,wpos)	Setzt den Schreibpointer nach wpos zurueck


Erweiterungen: DM 29.10.92
	slot= RB_PUSHSLOT(&rb)	gibt einen Verweis auf den naechsten
				freien Platz zurueck.
				Setzt voraus, dass der Puffer nicht voll
				ist (nicht getestet!)

	RB_PUSHADVANCE(&rb)	\"ubernimmt den naechsten freien Platz
				als Pufferinhalt.
				Setzt voraus, dass der Puffer nicht voll
				ist (nicht getestet!)

	slot= RB_POPSLOT(&rb)	gibt einen Verweis auf den naechsten
				belegten Platz zur\"uck.
				Setzt voraus, dass der Puffer nicht leer
				ist (nicht getestet!)

	RB_POPADVANCE(&rb)	streicht den naechsten belegten Platz
				als dem Puffer.
				Setzt voraus, dass der Puffer nicht leer
				ist (nicht getestet!)

Erweiterung: WS 19.02.93
	RB_CLEAN(&rb)		leert den Ringpuffer, 
				setzt in-Pointer = out-Pointer

Versionsgeschichte:

   DM	1.1	11.06.90
   DM	1.3	29.10.92	Ein-Ausgabezeiger jetzt aus Sicht des
  				Ringpuffers (nat\"urlicher!)

				Erweiterung f\"ur Daimler-Benz.

*/


#define RB_CREATE(rb,typ)	\
	struct {\
		typ	*rb_start; \
		typ	*rb_end; \
		typ	*rb_in; \
		typ	*rb_out; \
	} rb


#define RB_INIT(rb,start,anzahl) \
		( (rb)->rb_in= (rb)->rb_out= (rb)->rb_start= start,\
		  (rb)->rb_end= &(rb)->rb_start[anzahl]\
		)

#define RB_SLOT(rb,slot) \
		((slot)==(rb)->rb_end? (rb)->rb_start: (slot))
#define RB_SLOTBEG(rb,slot) \
		((slot) < (rb)->rb_start? (rb)->rb_end-1: (slot))

#define RB_EMPTY(rb)	((rb)->rb_in==(rb)->rb_out)

#define RB_FULL(rb)	(RB_SLOT(rb,(rb)->rb_in+1)==(rb)->rb_out)

#define RB_PUSH(rb,v) \
	(RB_FULL(rb) ? 1 : \
	 (*RB_PUSHSLOT(rb)= (v), RB_PUSHADVANCE(rb), 0) \
	)

#define RB_PUSHMINUS(rb,v) \
	(RB_EMPTY(rb) ? 1: \
	 ( (rb)->rb_in = RB_SLOTBEG((rb), (rb)->rb_in-1), *v = *(rb)->rb_in, 0)\
	)

#define RB_POP(rb,v) \
	(RB_EMPTY(rb) ? 1: \
	 (*v = *RB_POPSLOT(rb), RB_POPADVANCE(rb), 0) \
	)

#define RB_WPOS(rb) \
	((rb)->rb_in)

#define RB_WRESET(rb,wpos) \
	((rb)->rb_in = wpos)

#define RB_PUSHSLOT(rb) ((rb)->rb_in)
#define RB_PUSHADVANCE(rb) ((rb)->rb_in= RB_SLOT((rb), (rb)->rb_in+1))

#define RB_POPSLOT(rb) ((rb)->rb_out)
#define RB_POPADVANCE(rb) ((rb)->rb_out= RB_SLOT((rb), (rb)->rb_out+1))

#define RB_CLEAN(rb) ((rb)->rb_out=(rb)->rb_in)
