/*
** serial.c -- serial port support for SAB C16x
**
** Copyright (c) 1991-94 Hightec EDV-Systeme GmbH  D-66386 St.Ingbert
**
**      @(#) serial.c  3.0  01/28/94  
*/

/* this module uses two interrupt service routines for I/O :
	SerOutchar_Hnd ()
           on Trap 0x2d for SER1
		or
           on Trap 0x2a for SER0
   and SerInchar_Hnd ()
 	   on Trap 0x2e for SER1
		or 
           on Trap 0x2b for SER0
   Only one (exclusively) Serial port can be used at a time, 'cuz
   GDB166 - Target Monitor communication normally uses the other one !!
   Via macro SER0 one can select the serial port


*/

/*****************************************************************************
 * Includes
 *****************************************************************************/

#include <c166-ali.h>
#include <std.h>
#include <stdio.h>
#include <gnutrap.h>

#include "serial.h"
#include "rb.h"                 /* RingBuffer Implementierung           */

#define MAX_BUFLEN  128

static char outbuf[2*MAX_BUFLEN], inbuf[2*MAX_BUFLEN];
static int	active = 0;     /* zeigt an ob Ausgabe aktiv */

static unsigned int volatile	*outadr, *inadr;

static RB_CREATE(volatile out,char );     /* kreiert Ringbuffer fuer Ausgabe */
static RB_CREATE(volatile in,char ); 	/* kreiert Ringbuffer fuer Eingabe */
static unsigned int volatile * TIC_p;


/*****************************************************************************
	C166 specific IO-Macros fuer optimierte Word-Zugriffe unter 
        Voraussetzung, dass das Objekt im internen RAM liegt 
*****************************************************************************/

#define SET_INTERNAL_WORD(addr,val)	\
			asm volatile("mov r4, %0; or r4,#0xc000; mov [r4],%1" \
							:: "r"((int)addr),"r"((int)val) \
							: "r4");

#define GET_INTERNAL_WORD(addr)({ \
			int __X; \
			asm volatile("mov %0,%1; or %0,#0xc000; mov %0,[%0]" \
							: "=&r"(__X) \
							: "r" ((int)addr)); \
			__X;})

/*----------------------------------------------------------------------------* 
 * Handler (Interruptroutinen) der seriellen Schnittstelle                    * 
 *----------------------------------------------------------------------------*/

/*
 * Transmit Handler
 */

#if defined SER0
#define SEROUT_TRAP  0x2a
#define SERIN_TRAP   0x2b
#else
#define SEROUT_TRAP  0x2d
#define SERIN_TRAP   0x2e
#endif

/* declare SerOutchar_Hnd as reeentrant interrupt on Trap #SEROUT_TRAP, not
   saving/restoring multiply/division hardware
*/

TRAP_noMD (SEROUT_TRAP, SerOutchar_Hnd);

void
SerOutchar_Hnd(void)
{
	if(!RB_EMPTY(&out)){
		SET_INTERNAL_WORD(outadr,*RB_POPSLOT(&out));
		RB_POPADVANCE(&out);
	}
	else {
		active = 0;   /* Keine Ausgabe aktiv */
	}
}

/*
 * Receive Handler
 */

/* declare SerInchar_Hnd as reeentrant interrupt on Trap #SERIN_TRAP, not
   saving/restoring multiply/division hardware
*/

TRAP_noMD (SERIN_TRAP, SerInchar_Hnd);

void
SerInchar_Hnd (void)
{
	if(!RB_FULL(&in)){
		RB_PUSH(&in, GET_INTERNAL_WORD(inadr));
	}
}

/*----------------------------------------------------------------------------* 
 * Kommunikationsfunktionen                                                   *
 *----------------------------------------------------------------------------*/

/*
 * Ausgabe-Funktion fuer einzelne Zeichen
 */

int
serOutchar(char c)
{
	while  (RB_FULL(&out)) {
	  ; /* waiting */
	}
	RB_PUSH(&out,c);

	if(!active){
		active = 1;
		SET_INTERNAL_WORD(TIC_p,GET_INTERNAL_WORD(TIC_p) | 0x80);
	}

	return(0);
}

/*
 * Lese-Funktion
 */

char
serInchar(void)
{
	char c;

	 while (RB_POP(&in,&c)){
	   ;    /* waiting */
	}
	return(c);
}

int 
serOutstring(char *buf)
{
	unsigned char	len;

	for ( len = 0; len < strlen(buf); len++ ) {
		serOutchar(buf[len]);
	}
	return (len);
}

void
serInit(enum SERPORTS port, long baudrate, long hertz) 
{
	long baud = 1000000L*hertz;
       RB_INIT(&out,outbuf,256);
       RB_INIT(&in,inbuf,256);

	baud /= (baudrate*64);

	SET_SFRBIT(p3.8);
	SET_SFRBIT(dp3.8);
	CLR_SFRBIT(dp3.9);

	if (port == SER_1) {
                                /* Kommunikation ueber die
				       serielle Schnittstelle 1 */
	S1BG = (int) --baud;
	S1CON = 0x8011;

	inadr = &S1RBUF;     	/* Benutzung von Serial Port1 */
	outadr = &S1TBUF;


	TIC_p = &S1TIC;
	S1TIC = 0x49;
	S1RIC = 0x4a;
       }
	else  if (port == SER_0)  {
                                /* Kommunikation ueber die
				       serielle Schnittstelle 0 */
	S0BG = (int) --baud;
	S0CON = 0x8011;

	inadr = &S0RBUF;     	/* Benutzung von Serial Port0 */
	outadr = &S0TBUF;

	TIC_p = &S0TIC;
	S0TIC = 0x49;
	S0RIC = 0x4a;
      }

}
/******************************* EOF *****************************************/
